#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
from datetime import timedelta

import numpy as np
import pandas as pd
from pandas.tseries.offsets import MonthEnd
import pytz

from .exceptions import NoBaselineDataError, NoReportingDataError
from .warnings import EEMeterWarning

__all__ = (
    "Term",
    "as_freq",
    "day_counts",
    "get_baseline_data",
    "get_reporting_data",
    "get_terms",
    "remove_duplicates",
    "overwrite_partial_rows_with_nan",
    "clean_caltrack_billing_data",
    "clean_caltrack_billing_daily_data",
    "add_freq",
    "trim",
    "format_energy_data_for_caltrack",
    "format_temperature_data_for_caltrack",
)


def overwrite_partial_rows_with_nan(df):
    return df.dropna().reindex(df.index)


def remove_duplicates(df_or_series):
    """Remove duplicate rows or values by keeping the first of each duplicate.

    Parameters
    ----------
    df_or_series : :any:`pandas.DataFrame` or :any:`pandas.Series`
        Pandas object from which to drop duplicate index values.

    Returns
    -------
    deduplicated : :any:`pandas.DataFrame` or :any:`pandas.Series`
        The deduplicated pandas object.
    """
    # CalTrack 2.3.2.2
    return df_or_series[~df_or_series.index.duplicated(keep="first")]


def as_freq(
    data_series,
    freq,
    atomic_freq="1 Min",
    series_type="cumulative",
    include_coverage=False,
):
    """Resample data to a different frequency.

    This method can be used to upsample or downsample meter data. The
    assumption it makes to do so is that meter data is constant and averaged
    over the given periods. For instance, to convert billing-period data to
    daily data, this method first upsamples to the atomic frequency
    (1 minute freqency, by default), "spreading" usage evenly across all
    minutes in each period. Then it downsamples to hourly frequency and
    returns that result. With instantaneous series, the data is copied to all
    contiguous time intervals and the mean over `freq` is returned.

    **Caveats**:

     - This method gives a fair amount of flexibility in
       resampling as long as you are OK with the assumption that usage is
       constant over the period (this assumption is generally broken in
       observed data at large enough frequencies, so this caveat should not be
       taken lightly).

    Parameters
    ----------
    data_series : :any:`pandas.Series`
        Data to resample. Should have a :any:`pandas.DatetimeIndex`.
    freq : :any:`str`
        The frequency to resample to. This should be given in a form recognized
        by the :any:`pandas.Series.resample` method.
    atomic_freq : :any:`str`, optional
        The "atomic" frequency of the intermediate data form. This can be
        adjusted to a higher atomic frequency to increase speed or memory
        performance.
    series_type : :any:`str`, {'cumulative', ‘instantaneous’},
        default 'cumulative'
        Type of data sampling. 'cumulative' data can be spread over smaller
        time intervals and is aggregated using addition (e.g. meter data).
        'instantaneous' data is copied (not spread) over smaller time intervals
        and is aggregated by averaging (e.g. weather data).
    include_coverage: :any:`bool`,
        default `False`
        Option of whether to return a series with just the resampled values
        or a dataframe with a column that includes percent coverage of source data
        used for each sample.

    Returns
    -------
    resampled_data : :any:`pandas.Series` or :any:`pandas.DataFrame`
        Data resampled to the given frequency (optionally as a dataframe with a coverage column if `include_coverage` is used.
    """
    # TODO(philngo): make sure this complies with CalTRACK 2.2.2.1
    if not isinstance(data_series, pd.Series):
        raise ValueError(
            "expected series, got object with class {}".format(data_series.__class__)
        )
    if data_series.empty:
        return data_series
    series = remove_duplicates(data_series)
    target_freq = pd.Timedelta(atomic_freq)
    timedeltas = (series.index[1:] - series.index[:-1]).append(
        pd.TimedeltaIndex([pd.NaT])
    )

    if series_type == "cumulative":
        spread_factor = target_freq.total_seconds() / timedeltas.total_seconds()
        series_spread = series * spread_factor
        atomic_series = series_spread.asfreq(atomic_freq, method="ffill")
        resampled = atomic_series.resample(freq).sum()
        resampled_with_nans = atomic_series.resample(freq).first()
        n_coverage = atomic_series.resample(freq).count()
        resampled = resampled[resampled_with_nans.notnull()].reindex(resampled.index)

    elif series_type == "instantaneous":
        atomic_series = series.asfreq(atomic_freq, method="ffill")
        resampled = atomic_series.resample(freq).mean()

    if resampled.index[-1] < series.index[-1]:
        # this adds a null at the end using the target frequency
        last_index = pd.date_range(resampled.index[-1], freq=freq, periods=2)[1:]
        resampled = (
            pd.concat([resampled, pd.Series(np.nan, index=last_index)])
            .resample(freq)
            .mean()
        )
    if include_coverage:
        n_total = resampled.resample(atomic_freq).count().resample(freq).count()
        resampled = resampled.to_frame("value")
        resampled["coverage"] = n_coverage / n_total
        return resampled
    else:
        return resampled


def day_counts(index):
    """Days between DatetimeIndex values as a :any:`pandas.Series`.

    Parameters
    ----------
    index : :any:`pandas.DatetimeIndex`
        The index for which to get day counts.

    Returns
    -------
    day_counts : :any:`pandas.Series`
        A :any:`pandas.Series` with counts of days between periods. Counts are
        given on start dates of periods.
    """
    # dont affect the original data
    index = index.copy()

    if len(index) == 0:
        return pd.Series([], index=index)

    timedeltas = (index[1:] - index[:-1]).append(pd.TimedeltaIndex([pd.NaT]))
    timedelta_days = timedeltas.total_seconds() / (60 * 60 * 24)

    return pd.Series(timedelta_days, index=index)


def _make_baseline_warnings(
    end_inf, start_inf, data_start, data_end, start_limit, end_limit
):
    warnings = []
    # warn if there is a gap at end
    if not end_inf and data_end < end_limit:
        warnings.append(
            EEMeterWarning(
                qualified_name="eemeter.get_baseline_data.gap_at_baseline_end",
                description=(
                    "Data does not have coverage at requested baseline end date."
                ),
                data={
                    "requested_end": end_limit.isoformat(),
                    "data_end": data_end.isoformat(),
                },
            )
        )
    # warn if there is a gap at start
    if not start_inf and start_limit < data_start:
        warnings.append(
            EEMeterWarning(
                qualified_name="eemeter.get_baseline_data.gap_at_baseline_start",
                description=(
                    "Data does not have coverage at requested baseline start date."
                ),
                data={
                    "requested_start": start_limit.isoformat(),
                    "data_start": data_start.isoformat(),
                },
            )
        )
    return warnings


def get_baseline_data(
    data,
    start=None,
    end=None,
    max_days=365,
    allow_billing_period_overshoot=False,
    n_days_billing_period_overshoot=None,
    ignore_billing_period_gap_for_day_count=False,
):
    """Filter down to baseline period data.

    .. note::

        For compliance with CalTRACK, set ``max_days=365`` (section 2.2.1.1).

    Parameters
    ----------
    data : :any:`pandas.DataFrame` or :any:`pandas.Series`
        The data to filter to baseline data. This data will be filtered down
        to an acceptable baseline period according to the dates passed as
        `start` and `end`, or the maximum period specified with `max_days`.
    start : :any:`datetime.datetime`
        A timezone-aware datetime that represents the earliest allowable moment
        for the baseline data. The stricter of this or `max_days` is used
        to determine the earliest allowable baseline period timestamp.
    end : :any:`datetime.datetime`
        A timezone-aware datetime that represents the latest allowable end
        moment for the baseline data, i.e., the latest moment for which data is
        available before the intervention begins.
    max_days : :any:`int`, default 365
        The maximum length of the period. Ignored if `end` is not set.
        The stricter of this or `start` is used to determine the earliest
        allowable moment for the baseline period.
    allow_billing_period_overshoot : :any:`bool`, default False
        If True, count `max_days` from the end of the last billing data period
        that ends before the `end` date, rather than from the exact `end` date.
        Otherwise use the exact `end` date as the cutoff.
    n_days_billing_period_overshoot: :any:`int`, default None
        If `allow_billing_period_overshoot` is set to True, this determines
        the number of days of overshoot that will be tolerated. A value of
        None implies that any number of days is allowed.
    ignore_billing_period_gap_for_day_count : :any:`bool`, default False
        If True, instead of going back `max_days` from either the
        `end` date or end of the last billing period before that date (depending
        on the value of the `allow_billing_period_overshoot` setting) and
        excluding the last period that began before that date, first check to
        see if excluding or including that period gets closer to a total of
        `max_days` of data.

        For example, with `max_days=365`, if an exact 365 period would targeted
        Feb 15, but the billing period went from Jan 20 to Feb 20, exclude that
        period for a total of ~360 days of data, because that's closer to 365
        than ~390 days, which would be the total if that period was included.
        If, on the other hand, if that period started Feb 10 and went to Mar 10,
        include the period, because ~370 days of data is closer to than ~340.

    Returns
    -------
    baseline_data, warnings : :any:`tuple` of (:any:`pandas.DataFrame` or :any:`pandas.Series`, :any:`list` of :any:`eemeter.EEMeterWarning`)
        Data for only the specified baseline period and any associated warnings.
    """
    if max_days is not None:
        if start is not None:
            raise ValueError(  # pragma: no cover
                "If max_days is set, start cannot be set: start={}, max_days={}.".format(
                    start, max_days
                )
            )

    start_inf = False
    if start is None:
        # py datetime min/max are out of range of pd.Timestamp min/max
        start_target = pytz.UTC.localize(pd.Timestamp.min) + timedelta(days=1)
        start_inf = True
    else:
        start_target = start

    end_inf = False
    if end is None:
        end_limit = pytz.UTC.localize(pd.Timestamp.max) - timedelta(days=1)
        end_inf = True
    else:
        end_limit = end

    # copying prevents setting on slice warnings
    data_before_end_limit = data[:end_limit].copy()
    data_end = data_before_end_limit.index.max()

    if ignore_billing_period_gap_for_day_count and (
        n_days_billing_period_overshoot is None
        or end_limit - timedelta(days=n_days_billing_period_overshoot) < data_end
    ):
        end_limit = data_before_end_limit.index.max()

    if not end_inf and max_days is not None:
        start_target = end_limit - timedelta(days=max_days)

    if allow_billing_period_overshoot:
        # adjust start limit to get a selection closest to max_days
        # also consider ffill for get_loc method - always picks previous
        try:
            loc = data_before_end_limit.index.get_indexer(
                [start_target], method="nearest"
            )[0]
        except (KeyError, IndexError):  # pragma: no cover
            baseline_data = data_before_end_limit
            start_limit = start_target
        else:
            start_limit = data_before_end_limit.index[loc]
            baseline_data = data_before_end_limit[start_limit:].copy()

    else:
        # use hard limit for baseline start
        start_limit = start_target
        baseline_data = data_before_end_limit[start_limit:].copy()

    if baseline_data.dropna().empty:
        raise NoBaselineDataError()

    baseline_data.iloc[-1] = np.nan

    data_end = data.index.max()
    data_start = data.index.min()
    return (
        baseline_data,
        _make_baseline_warnings(
            end_inf, start_inf, data_start, data_end, start_limit, end_limit
        ),
    )


def _make_reporting_warnings(
    end_inf, start_inf, data_start, data_end, start_limit, end_limit
):
    warnings = []
    # warn if there is a gap at end
    if not end_inf and data_end < end_limit:
        warnings.append(
            EEMeterWarning(
                qualified_name="eemeter.get_reporting_data.gap_at_reporting_end",
                description=(
                    "Data does not have coverage at requested reporting end date."
                ),
                data={
                    "requested_end": end_limit.isoformat(),
                    "data_end": data_end.isoformat(),
                },
            )
        )
    # warn if there is a gap at start
    if not start_inf and start_limit < data_start:
        warnings.append(
            EEMeterWarning(
                qualified_name="eemeter.get_reporting_data.gap_at_reporting_start",
                description=(
                    "Data does not have coverage at requested reporting start date."
                ),
                data={
                    "requested_start": start_limit.isoformat(),
                    "data_start": data_start.isoformat(),
                },
            )
        )
    return warnings


def get_reporting_data(
    data,
    start=None,
    end=None,
    max_days=365,
    allow_billing_period_overshoot=False,
    ignore_billing_period_gap_for_day_count=False,
):
    """Filter down to reporting period data.

    Parameters
    ----------
    data : :any:`pandas.DataFrame` or :any:`pandas.Series`
        The data to filter to reporting data. This data will be filtered down
        to an acceptable reporting period according to the dates passed as
        `start` and `end`, or the maximum period specified with `max_days`.
    start : :any:`datetime.datetime`
        A timezone-aware datetime that represents the earliest allowable moment
        for the reporting data, i.e., the earliest moment for which data is
        available after the intervention begins.
    end : :any:`datetime.datetime`
        A timezone-aware datetime that represents the latest allowable end
        moment for the reporting data. The stricter of this or `max_days` is used
        to determine the latest allowable reporting period timestamp.
    max_days : :any:`int`, default 365
        The maximum length of the period. Ignored if `start` is not set.
        The stricter of this or `end` is used to determine the latest
        allowable reporting period timestamp.
    allow_billing_period_overshoot : :any:`bool`, default False
        If True, count `max_days` from the start of the first billing data period
        that starts after the `start` date, rather than from the exact `start` date.
        Otherwise use the exact `start` date as the cutoff.
    ignore_billing_period_gap_for_day_count : :any:`bool`, default False
        If True, instead of going forward `max_days` from either the
        `start` date or the `start` of the first billing period after that date
        (depending on the value of the `allow_billing_period_overshoot` setting)
        and excluding the first period that ended after that date, first check
        to see if excluding or including that period gets closer to a total of
        `max_days` of data.

        For example, with `max_days=365`, if an exact 365 period would targeted
        Feb 15, but the billing period went from Jan 20 to Feb 20, include that
        period for a total of ~370 days of data, because that's closer to 365
        than ~340 days, which would be the total if that period was excluded.
        If, on the other hand, if that period started Feb 10 and went to Mar 10,
        exclude the period, because ~360 days of data is closer to than ~390.

    Returns
    -------
    reporting_data, warnings : :any:`tuple` of (:any:`pandas.DataFrame` or
    :any:`pandas.Series`, :any:`list` of :any:`eemeter.EEMeterWarning`)
        Data for only the specified reporting period and any associated warnings.
    """
    if max_days is not None:
        if end is not None:
            raise ValueError(  # pragma: no cover
                "If max_days is set, end cannot be set: end={}, max_days={}.".format(
                    end, max_days
                )
            )

    start_inf = False
    if start is None:
        # py datetime min/max are out of range of pd.Timestamp min/max
        start_limit = pytz.UTC.localize(pd.Timestamp.min) + timedelta(days=1)
        start_inf = True
    else:
        start_limit = start

    end_inf = False
    if end is None:
        end_target = pytz.UTC.localize(pd.Timestamp.max) - timedelta(days=1)
        end_inf = True
    else:
        end_target = end

    # copying prevents setting on slice warnings
    data_after_start_limit = data[start_limit:].copy()

    if ignore_billing_period_gap_for_day_count:
        start_limit = data_after_start_limit.index.min()

    if not start_inf and max_days is not None:
        end_target = start_limit + timedelta(days=max_days)

    if allow_billing_period_overshoot:
        # adjust start limit to get a selection closest to max_days
        # also consider bfill for get_loc method - always picks next
        try:
            loc = data_after_start_limit.index.get_indexer(
                [end_target], method="nearest"
            )[0]
        except (KeyError, IndexError):  # pragma: no cover
            reporting_data = data_after_start_limit
            end_limit = end_target
        else:
            end_limit = data_after_start_limit.index[loc]
            reporting_data = data_after_start_limit[:end_limit].copy()

    else:
        # use hard limit for baseline start
        end_limit = end_target
        reporting_data = data_after_start_limit[:end_limit].copy()

    if reporting_data.dropna().empty:
        raise NoReportingDataError()

    reporting_data.iloc[-1] = np.nan

    data_end = data.index.max()
    data_start = data.index.min()
    return (
        reporting_data,
        _make_reporting_warnings(
            end_inf, start_inf, data_start, data_end, start_limit, end_limit
        ),
    )


class Term(object):
    """
    The term object represents a subset of an index.

    Attributes
    ----------
    index : :any:`pandas.DatetimeIndex`
        The index of the term. Includes a period at the end meant to be NaN-value.
    label : :any:`str`
        The label for the term.
    target_start_date : :any:`pandas.Timestamp` or :any:`datetime.datetime`
        The start date inferred for this term from the start date and target term
        lenths.
    target_end_date : :any:`pandas.Timestamp` or :any:`datetime.datetime`
        The end date inferred for this term from the start date and target term
        lenths.
    target_term_length_days : :any:`int`
        The number of days targeted for this term.
    actual_start_date : :any:`pandas.Timestamp`
        The first date in the index.
    actual_end_date : :any:`pandas.Timestamp`
        The last date in the index.
    actual_term_length_days : :any:`int`
        The number of days between the actual start date and actual end date.
    complete : :any:`bool`
        True if this term is conclusively complete, such that additional data added
        to the series would not add more data to this term.

    """

    def __init__(
        self,
        index,
        label,
        target_start_date,
        target_end_date,
        target_term_length_days,
        actual_start_date,
        actual_end_date,
        actual_term_length_days,
        complete,
    ):
        self.index = index
        self.label = label
        self.target_start_date = target_start_date
        self.target_end_date = target_end_date
        self.target_term_length_days = target_term_length_days
        self.actual_start_date = actual_start_date
        self.actual_end_date = actual_end_date
        self.actual_term_length_days = actual_term_length_days
        self.complete = complete

    def __repr__(self):
        return (
            "Term(label={}, target_term_length_days={}, actual_term_length_days={},"
            " complete={})"
        ).format(
            self.label,
            self.target_term_length_days,
            self.actual_term_length_days,
            self.complete,
        )


def get_terms(index, term_lengths, term_labels=None, start=None, method="strict"):
    """Breaks a :any:`pandas.DatetimeIndex` into consecutive terms of specified
    lengths.

    Parameters
    ----------
    index : :any:`pandas.DatetimeIndex`
        The index to split into terms, generally `meter_data.index`
        or `temperature_data.index`.
    term_lengths : :any:`list` of :any:`int`
        The lengths (in days) of the terms into which to split the data.
    term_labels : :any:`list` of :any:`str`, default None
        Labels to use for each term. List must be the same length as the
        `term_lengths` list.
    start : :any:`datetime.datetime`, default None
        A timezone-aware datetime that represents the earliest allowable start
        date for the terms. If None, use the first element of the index.
    method: one of ['strict', 'nearest'], default 'strict'
        The method to use to get terms.

        - "strict": Ensures that the term end will come on or before the length of

    Returns
    -------
    terms : :any:`list` of :any:`eemeter.Term`
        A dataframe of term labels with the same :any:`pandas.DatetimeIndex`
        given as `index`. This can be used to filter the original data into
        terms of approximately the desired length.


    """
    if method == "strict":
        get_loc_method = "pad"
    elif method == "nearest":
        get_loc_method = "nearest"
    else:
        raise ValueError(
            "method {} not supported - use either 'strict' or 'closest'".format(method)
        )

    if not index.is_monotonic_increasing:
        raise ValueError("get_terms requires a sorted index")

    if term_labels is None:
        term_labels = [
            "term_{:03d}".format(i + 1) for i, term_length in enumerate(term_lengths)
        ]

    elif len(term_labels) != len(term_lengths):
        raise ValueError(
            "term_labels (len {}) must be the same length as term_length (len {})".format(
                len(term_labels), len(term_lengths)
            )
        )

    if start is None:
        prev_start = index.min()
    else:
        prev_start = start

    term_end_targets = [
        prev_start + timedelta(days=sum(term_lengths[: i + 1]))
        for i in range(len(term_lengths))
    ]

    terms = []
    remaining_index = index[index >= prev_start]

    for label, target_term_length, end_target in zip(
        term_labels, term_lengths, term_end_targets
    ):
        if len(remaining_index) <= 1:
            break

        next_index = remaining_index.get_indexer([end_target], method=get_loc_method)[0]

        # keep one extra index point for the end NaN - this could be confusing, but
        # helps identify the full range of the last data point
        term_index = remaining_index[: next_index + 1]

        # find the next start
        next_start = remaining_index[next_index]

        # reset the remaining index
        remaining_index = remaining_index[next_index:]

        # There may be a better way to tell if the term is conclusively complete,
        # but the logic here is that if there's more than one remaining point then
        # the term must be complete - since that final point was a worse candidate
        # than the one before it which was chosen.
        complete = len(remaining_index) > 1

        terms.append(
            Term(
                index=term_index,
                label=label,
                target_start_date=prev_start,
                target_end_date=end_target,
                target_term_length_days=target_term_length,
                actual_start_date=term_index[0],
                actual_end_date=term_index[-1],
                actual_term_length_days=(term_index[-1] - term_index[0]).days,
                complete=complete,
            )
        )

        # reset the previous start
        prev_start = next_start

    return terms


def clean_caltrack_billing_data(data, source_interval):
    # check for empty data
    if data["value"].dropna().empty:
        return data[:0]

    if source_interval.startswith("billing"):
        diff = list((data.index[1:] - data.index[:-1]).days)
        filter_ = pd.Series(diff + [np.nan], index=data.index)

        # CalTRACK 2.2.3.4, 2.2.3.5
        if source_interval == "billing_monthly":
            data = data[
                (filter_ <= 35) & (filter_ >= 25)  # keep these, inclusive
            ].reindex(data.index)

        # CalTRACK 2.2.3.4, 2.2.3.5
        if source_interval == "billing_bimonthly":
            data = data[
                (filter_ <= 70) & (filter_ >= 25)  # keep these, inclusive
            ].reindex(data.index)

        # CalTRACK 2.2.3.1
        """
        Adds estimate to subsequent read if there aren't more than one estimate in a row
        and then removes the estimated row.

        Input:
        index   value   estimated
        1       2       False
        2       3       False
        3       5       True
        4       4       False
        5       6       True
        6       3       True
        7       4       False
        8       NaN     NaN

        Output:
        index   value
        1       2
        2       3
        4       9
        5       NaN
        7       7
        8       NaN
        """
        add_estimated = []
        remove_estimated_fixed_rows = []
        orig_data = data.copy()
        if "estimated" in data.columns:
            data["unestimated_value"] = (
                data[:-1].value[(data[:-1].estimated == False)].reindex(data.index)
            )
            data["estimated_value"] = (
                data[:-1].value[(data[:-1].estimated)].reindex(data.index)
            )
            for i, (index, row) in enumerate(data[:-1].iterrows()):
                # ensures there is a prev_row and previous row value is null
                if i > 0 and pd.isnull(prev_row["unestimated_value"]):
                    # current row value is not null
                    add_estimated.append(prev_row["estimated_value"])
                    if not pd.isnull(row["unestimated_value"]):
                        # get all rows that had only estimated reads that will be
                        # added to the subsequent row meaning this row
                        # needs to be removed
                        remove_estimated_fixed_rows.append(prev_index)
                else:
                    add_estimated.append(0)
                prev_row = row
                prev_index = index
            add_estimated.append(np.nan)
            data["value"] = data["unestimated_value"] + add_estimated
            data = data[~data.index.isin(remove_estimated_fixed_rows)]
            data = data[["value"]]  # remove the estimated column

    # check again for empty data
    if data.dropna().empty:
        return data[:0]

    return data


def downsample_and_clean_caltrack_daily_data(data):
    data = as_freq(data.value, "D", include_coverage=True)

    # CalTRACK 2.2.2.1 - interpolate with average of non-null values
    data.loc[data.coverage > 0.5, "value"] = (
        data[data.coverage > 0.5].value / data[data.coverage > 0.5].coverage
    )

    # CalTRACK 2.2.2.1 - discard days with less than 50% coverage
    return data[data.coverage > 0.5].reindex(data.index)[["value"]]


def clean_caltrack_billing_daily_data(data, source_interval):
    # billing data is cleaned but not resampled
    if source_interval.startswith("billing"):
        # CalTRACK 2.2.3.4, 2.2.3.5
        return clean_caltrack_billing_data(data, source_interval)

    # higher intervals like daily, hourly, 30min, 15min are
    # resampled (daily) or downsampled (hourly, 30min, 15min)
    elif source_interval == "daily":
        return data
    else:
        return downsample_and_clean_caltrack_daily_data(data)


def add_freq(idx, freq=None):
    """Add a frequency attribute to idx, through inference or directly.

     Returns a copy.  If `freq` is None, it is inferred.

     Note: this function is taken from
     https://stackoverflow.com/questions/46217529/pandas-datetimeindex-frequency-is-none-and-cant-be-set;
     credit Brad Solomon.


    Parameters
    ----------
    idx : :any:`pandas.DateTimeIndex`
        Any DateTimeIndex.
    freq : :any valid DateTimeIndex Freq in 'str' format
        The frequency of the datetime index. Defaults to 'None' if frequency is to be inferred.

    Returns
    -------
    idx : :any:`pandas.DateTimeIndex`
        A copy of idx with frequency added.
    """

    idx = idx.copy()
    if freq is None:
        if idx.freq is None:
            freq = pd.infer_freq(idx)
        else:
            return idx
    idx.freq = pd.tseries.frequencies.to_offset(freq)
    if idx.freq is None:
        raise AttributeError(
            "no discernible frequency found to `idx`.  Specify"
            " a frequency string with `freq`."
        )
    return idx


def trim(*args, freq="h", tz="UTC"):
    """A helper function which trims a given number of time series dataframes so that they all correspond to the same
    time periods. Typically used to ensure that both gas, electricity, and temperature datasets cover the same time
    period. Trim undertakes the following steps:

       - copies dataframes
       - sets indexes to datetimes if not already
       - localises index to UTC (default - if other timezone applies this should be specified)
       - sorts in ascending order against DateTimeIndex
       - drops nulls at both start and end of df
       - trims the dataframes by equalising all min(df.index) and max(df.index)

       Trim requires both input dataframes to have some degree of overlap beforehand.

     Parameters
    ----------
    *args : : one or more 'pandas.DataFrame's
        A set of regular time series datasets. If index is not DateTimeIndex, function will convert accordingly. There
        must be overlap between all datasets otherwise trim will return IndexError. Can function with one dataframe,
        though not much point given functionality.
    freq : : any valid DateTimeIndex frequency.
        This is used to identify any duplicates and missing values in a dataframe and ensure that each dataframe in the
        returned tuple is of the same length. Freq defaults to '1H' (one hour) but can be, for example '0.5H' (1/2 hour)
    tz : : any valid timezone 'str'
        The timezone associated with the given dataframes. If timezone-naive, function will localise to 'UTC' as
        default.

     Returns
    -------
     out_dfs : :any:`tuple` of 'pandas.DataFrame's.
         A list of dataframes trimmed to equal total intervals, arranged in eemeter format (i.e. with ascending
         indices).
    """
    new_tuple = ()
    if len(list(args)) == 1:
        args = args[0]
    for i in args:
        df = i
        if not isinstance(df.index, pd.DatetimeIndex):
            df.index = pd.to_datetime(df.index, infer_datetime_format=True)
        if df.index.tz is None:
            df.index = df.index.tz_localize(tz=tz)  # defaults to UTC
        df = df.sort_index()
        df = df[~df.index.duplicated(keep="first")]
        df = df.resample(freq).asfreq()
        new_tuple = new_tuple + (df,)
    max_start = max([min(df.index) for df in new_tuple])
    min_end = min([max(df.index) for df in new_tuple])
    out_dfs = ()
    if max_start < min_end:
        for df in new_tuple:
            out_dfs = out_dfs + (df.loc[max_start:min_end],)
    else:
        raise IndexError("Trim requires for all dfs to have some overlap.")

    return out_dfs


def _check_input_formatting(input, tz="UTC"):
    if not isinstance(input.index, pd.DatetimeIndex):
        if isinstance(input.index, pd.RangeIndex):
            for i in [
                "start",
                "Start",
                "Datetime",
                "timestamp",
                "Timestamp",
                "datetime",
            ]:  # this is a non-exhaustive list (welcome additions) of possible timestamp headers when not in index.
                if i in input.columns.values:
                    input = input.set_index(i)
        if not isinstance(input.index, pd.DatetimeIndex):
            input.index = pd.to_datetime(input.index)
            if input.index[0].tzinfo is None:
                input.index = input.index.tz_localize(tz=tz)
        else:
            raise ValueError(
                "Data is not in correct format - index should be of class 'pd.core.indexes.datetimes.DatetimeIndex',"
                + " or datetime column should be labelled one of: 'Start', 'start', 'Datetime', 'timestamp', "
                "'Timestamp', or 'Datetime'."
            )
    if input.index[0].tzinfo is None:
        input.index = input.index.tz_localize(tz=tz)
    return input


def _format_data_for_caltrack_hourly(df, tz="UTC"):
    if df is not None:
        df = df.copy()
        df = _check_input_formatting(df, tz)
        df = df.sort_index()
        return df
    else:
        return None


def format_energy_data_for_caltrack(*args, method="hourly", tz="UTC"):
    """A helper function which ensures energy consumption data is formatted for eemeter processing.

    Parameters
    ----------
    *args : :one or more `pandas.DataFrame`s
        Energy consumption time series data. Consumption must be measured in the same units.
    method : : any valid 'str'
        The relevant eemeter model requiring formatting. Must be either 'hourly', 'daily', or 'billing'. Defaults to
        'hourly'.
    tz : : any valid timezone 'str'
        The timezone associated with the given dataframes. If timezone-naive, function will localise to 'UTC' as
        default.

    Returns
    -------
    args_tuple : any 'list' containing one or more 'pandas.DataFrame's
        A list of dataframes comprising energy consumption data in eemeter format.
    """

    if method == "hourly":
        freq = "h"
    elif method == "daily":
        freq = "D"
    elif method == "billing":
        freq = MonthEnd()  # "M"/"ME" depending on pandas version
    else:
        raise ValueError("'method' must be either 'hourly', 'daily' or 'billing'.")

    args_tuple = ()
    for df in args:
        df = _format_data_for_caltrack_hourly(df, tz)
        if not isinstance(df, pd.DataFrame):
            df = pd.DataFrame(df)
        df = df.resample(freq).sum()
        df.index = df.index.rename("start")
        args_tuple = args_tuple + (df,)
        if df.columns[0] != "value":
            current_col_name = df.columns[0]
            df.rename(columns={current_col_name: "value"}, inplace=True)

    if len(args_tuple) == 1:
        return args_tuple[0]
    else:
        args_list = list(args_tuple)
        args_list[-1], args_list[-2] = trim(args_list[-1], args_list[-2], freq=freq)
        args_tuple = tuple(args_list)
        return args_tuple


def format_temperature_data_for_caltrack(temperature_data, tz="UTC"):
    """A helper function which ensures external temperature data is formatted for eemeter processing.

    Parameters
    ----------
    temperature_data : :any:``
        Hourly external temperature data. If DataFrame, not pd.Series (as required by CalTRACK) function will convert.
    tz : : any valid timezone 'str'
        The timezone associated with the given dataframes. If timezone-naive, function will localise to 'UTC' as
        default.
    Returns
    -------
    temperature_data : :any:``
        Hourly external temperature data in eemeter format.
    """

    temperature_data = _format_data_for_caltrack_hourly(temperature_data, tz)
    mask = temperature_data.index.minute == 00
    temperature_data = temperature_data[mask]
    if temperature_data.index.freq == None:
        temperature_data.index = add_freq(temperature_data.index)
    if isinstance(temperature_data, pd.DataFrame):
        temperature_data = temperature_data.squeeze()
    return temperature_data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

from pathlib import Path

import numpy as np
import pandas as pd
from opendsm.eemeter.models.daily.model import DailyModel
from opendsm.eemeter.models.daily.data import DailyBaselineData
from opendsm.eemeter.models.daily.optimize_results import OptimizedResult


# Define the current directory
current_dir = Path(__file__).resolve().parent


class TestFitModel:
    @classmethod
    def setup_class(cls):
        # Create a sample meter data DataFrame from the test data
        df = pd.read_csv(current_dir / "test_data.csv")
        df.index = pd.to_datetime(df["datetime"])
        df = df[["temperature", "observed"]]
        cls.meter_data = DailyBaselineData(df, is_electricity_data=True)

    def test_fit_model(self):
        # Create a DailyModel instance
        fm = DailyModel().fit(self.meter_data, ignore_disqualification=True)

        # Test that the combinations attribute is a list
        assert isinstance(fm.combinations, list)

        # Test that the combinations attribute is as expected
        expected_combinations = [
            "fw-su_sh_wi",
            "fw-sh_wi__fw-su",
            "fw-sh_wi__wd-su__we-su",
            "wd-su_sh_wi__we-su_sh_wi",
            "fw-su__wd-sh_wi__we-sh_wi",
            "wd-su__wd-sh_wi__we-su_sh_wi",
            "wd-su_sh_wi__we-su__we-sh_wi",
            "wd-su__wd-sh_wi__we-su__we-sh_wi",
        ]
        assert fm.combinations == expected_combinations

        # Test that the components attribute is a list
        assert isinstance(fm.components, list)

        # Test that the components attribute is as expected
        expected_components = [
            "fw-su",
            "wd-su",
            "we-su",
            "fw-sh_wi",
            "wd-sh_wi",
            "we-sh_wi",
            "fw-su_sh_wi",
            "wd-su_sh_wi",
            "we-su_sh_wi",
        ]
        assert fm.components == expected_components

        # Test that the fit_components attribute is a dictionary
        assert isinstance(fm.fit_components, dict)

        # Test that the wRMSE_base attribute is a float
        assert isinstance(fm.wRMSE_base, float)
        assert np.isclose(fm.wRMSE_base, 18.39, rtol=1e-2)

        # Test that the best combination is as expected
        expected_best_combination = "wd-su_sh_wi__we-su_sh_wi"
        assert fm.best_combination == expected_best_combination

        # Test that the final model is as expected
        combinations = expected_best_combination.split("__")
        for combination in combinations:
            assert isinstance(fm.model[combination], OptimizedResult)

        # Test that the error attribute values are as expected
        expected_model_error = {
            "wRMSE": 16.96,
            "RMSE": 16.96,
            "MAE": 13.40,
            "CVRMSE": 0.3207,
            "PNRMSE": 0.6326,
        }
        for k in expected_model_error:
            assert np.isclose(fm.error[k], expected_model_error[k], rtol=1e-2)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
import warnings

import numba
import numpy as np
import numpy.ma as ma

from sklearn.experimental import enable_iterative_imputer
from sklearn.impute import IterativeImputer
from sklearn.linear_model import BayesianRidge

from scipy.interpolate import RBFInterpolator

from copy import deepcopy as copy


def autocorr_fcn(x, lags, exclude_0=True):
    """manualy compute, non partial"""
    x_msk = ma.masked_invalid(x)
    mean = ma.mean(x_msk)
    var = ma.var(x_msk)
    xp = x_msk - mean
    corr = [1.0 if l == 0 else ma.sum(xp[l:] * xp[:-l]) / len(x) / var for l in lags]

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", "Warning: converting a masked element to nan")
        # combine the lags, the correlation values, and mirror to get leads/lags
        res = np.vstack((lags, corr)).T
    if exclude_0:  # remove the 0 lag
        res = res[1:]
        rev_res = copy(res)[::-1]
    else:
        rev_res = copy(res)[::-1][:-1]

    rev_res[:, 0] = -rev_res[:, 0]
    res = np.vstack((rev_res, res))

    return res


# unused
def autocorr_fcn2(x, lags):
    """np.correlate, non partial"""
    x_msk = ma.masked_invalid(x)
    mean = ma.mean(x_msk)
    var = ma.var(x_msk)
    xp = x_msk - mean

    corr = ma.correlate(xp, xp, "full")[len(x) - 1 :] / var / len(x)

    return corr[: len(lags)]


# unused
def autocorr_fcn3(x, lags):
    """fft, pad 0s, non partial"""
    x_msk = ma.masked_invalid(x)
    n = len(x)
    # pad 0s to 2n-1
    ext_size = 2 * n - 1
    # nearest power of 2
    fsize = 2 ** np.ceil(np.log2(ext_size)).astype("int")

    mean = ma.mean(x_msk)
    var = ma.var(x_msk)
    xp = x - mean

    # do fft and ifft
    cf = np.fft.fft(xp, fsize)
    sf = cf.conjugate() * cf
    corr = np.fft.ifft(sf).real
    corr = corr / var / n

    return corr[: len(lags)]


# unused
def multiple_imputation(df, columns=None, **kwargs):
    # get indices of missing values
    missing_idx = df[columns].isna().any(axis=1)

    df_imputed = df[columns].reset_index()
    # convert datetime to hours since earliest datetime
    df_imputed["datetime_elapsed"] = (
        df_imputed["datetime"] - df_imputed["datetime"].min()
    ).dt.total_seconds() / 3600
    df_imputed["hour_of_day"] = df_imputed["datetime"].dt.hour
    df_imputed["day_of_week"] = df_imputed["datetime"].dt.dayofweek
    df_imputed["month"] = df_imputed["datetime"].dt.month
    df_imputed = df_imputed.set_index("datetime")

    settings_dict = {
        "estimator": BayesianRidge(),  # can use SVR, BayesianRidge, etc.
        "max_iter": 10,
        "random_state": None,
    }
    settings_dict.update(kwargs)

    imputer = IterativeImputer(**settings_dict)
    imputer.fit(df_imputed)
    df_imputed[:] = imputer.transform(df_imputed)

    # add df_imputed back to df
    df.loc[missing_idx, columns] = df_imputed.loc[missing_idx, columns]

    # add additional columns to indicate which values were imputed
    for col in columns:
        interp_bool_col = f"interpolated_{col}"

        df[interp_bool_col] = False
        df.loc[missing_idx, interp_bool_col] = True

    return df


# @numba.njit
def shift_array(arr, num, fill_value=np.nan):
    # Courtesy of https://stackoverflow.com/questions/30399534/shift-elements-in-a-numpy-array
    # get size of arr
    arr_size = arr.shape[0]

    if arr_size <= 20000:
        if num >= 0:
            return np.concatenate((np.full(num, fill_value), arr[:-num]))
        else:
            return np.concatenate((arr[-num:], np.full(-num, fill_value)))
    else:
        result = np.empty_like(arr)

        if num > 0:
            result[:num] = fill_value
            result[num:] = arr[:-num]
        elif num < 0:
            result[num:] = fill_value
            result[:num] = arr[-num:]
        else:
            result[:] = arr

        return result


def _interpolate_col(x, lags):
    # check that the column has nans
    if x.isna().sum() == 0:
        return x

    elif x.isna().sum() == len(x):
        return x

    # calculate the number of lags and leads to consider
    if x.name == "observed":
        missing_frac = x.isna().sum() / len(x)
        n_cor_idx_heuristic = (
            np.round((4.012 * np.log(missing_frac) + 24.38) / 2, 0) * 2
        )
        n_cor_idx = int(np.max([6, n_cor_idx_heuristic]))
    else:
        n_cor_idx = 6

    # Calculate the correlation of col with its lags and leads
    # create lags from -lags to lags
    lag_array = np.arange(lags + 1)
    autocorr = autocorr_fcn(x.values, lag_array, exclude_0=True)

    # take the largest n_cor_idx from second column using argpartition
    idx = np.argpartition(autocorr[:, 1], -n_cor_idx)[-n_cor_idx:]
    autocorr = autocorr[idx]

    # sort autocorr by the autocorrelation value
    autocorr = autocorr[np.argsort(autocorr[:, 1])[::-1]]
    autocorr_idx = autocorr[:, 0]

    # interpolate and update the values
    max_iter = 10
    for i, cnt_min in enumerate(np.linspace(n_cor_idx, 1, max_iter).astype(int)):
        num_rows = x.shape[0]
        num_cols = len(autocorr_idx)

        autocorr_helpers = np.empty((num_rows, num_cols))
        for i in range(num_cols):
            shift = int(autocorr_idx[i])
            autocorr_helpers[:, i] = shift_array(x.values, shift)

        # get the indices of the missing values
        nan_series_idx = x.index[x.isna()]
        nan_idx = x.index.get_indexer(nan_series_idx)

        # nan values where helpers are not nan
        valid_idx = np.sum(~np.isnan(autocorr_helpers[nan_idx, :]), axis=1) >= cnt_min
        if valid_idx.sum() == 0:
            continue

        nan_series_idx = nan_series_idx[valid_idx]
        nan_idx = x.index.get_indexer(nan_series_idx)

        # for each row, if the value is missing, calculate the mean of the lags and leads
        # ignore FutureWarning from pandas for now
        with warnings.catch_warnings():
            warnings.simplefilter(action="ignore", category=FutureWarning)
            x.loc[nan_series_idx] = np.nanmean(autocorr_helpers[nan_idx, :], axis=1)

        if x.isna().sum() == 0:
            break

    return x


def interpolate(df, columns=None):
    skip_autocorr_interpolation = False
    if len(df) > 6 * 24 * 7:
        lags = 24 * 7 * 2 + 1
    elif (len(df) > 3 * 24 * 7) and (len(df) <= 6 * 24 * 7):
        lags = 24 * 7 + 1
    elif (len(df) > 3 * 24) and (len(df) <= 3 * 24 * 7):
        lags = 24 + 1
    else:
        skip_autocorr_interpolation = True

    interp_cols = columns
    if interp_cols is None:
        interp_cols = ["temperature", "ghi", "observed"]

    # check if the columns are in the dataframe and modify columns appropriately
    for col in interp_cols:
        if col not in df.columns:
            continue

        interp_bool_col = f"interpolated_{col}"
        if interp_bool_col in df.columns:
            continue

        # main interpolation method
        idx_missing = df.loc[df[col].isna()].index
        if not skip_autocorr_interpolation:
            df[col] = _interpolate_col(df[col].copy(), lags)

        # backup interpolation methods
        for method in ["time", "ffill", "bfill"]:
            na_datetime = df.loc[df[col].isna()].index
            if len(na_datetime) == 0:
                break

            if method == "time":
                df[col] = df[col].interpolate(method="time", limit_direction="both")

            elif method == "ffill":
                df[col] = df[col].ffill()

            elif method == "bfill":
                df[col] = df[col].bfill()

        # TODO: we can check if we have similar values multiple times back to back, if yes, raise a warning
        # where na_datetime_original is True and the col is not na, set the interpolation boolean to True
        df[interp_bool_col] = False
        df.loc[df.index.isin(idx_missing) & ~df[col].isna(), interp_bool_col] = True

    return df



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import colorsys
from copy import deepcopy as copy

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np

from opendsm.common.adaptive_loss import IQR_outlier
from opendsm.eemeter.models.daily.utilities.ellipsoid_test import (
    robust_confidence_ellipse,
)

fontsize = 14
mpl.rc("font", family="sans-serif")
c = ["tab:blue", "tab:green", "tab:purple"]


def adjust_lightness(color, amount=1.0):
    try:
        c = mpl.colors.cnames[color]
    except:
        c = color

    c = colorsys.rgb_to_hls(*mpl.colors.to_rgb(c))

    return colorsys.hls_to_rgb(c[0], max(0, min(1, amount * c[1])), c[2])


def plot(
    fit,
    meter_eval,
    include_resid=False,
    plot_gaussian_ellipses=False,
    plot_outliers=True,
):
    # sort meter_eval by temperature
    meter_eval = meter_eval.sort_values(by="temperature")
    meter_eval = meter_eval.dropna(subset=["temperature", "predicted"])

    fig = plt.figure(figsize=(14, 4), dpi=300)
    if include_resid:
        gs = fig.add_gridspec(2, hspace=0, height_ratios=[2.5, 1])
        ax = gs.subplots()
    else:
        ax = [fig.subplots()]

    # Plot scatter and Gaussian ellipses
    for n, season in enumerate(["summer", "shoulder", "winter"]):
        for day_type, day_num in enumerate([[0, 1, 2, 3, 4], [5, 6]]):
            if day_type == 0:
                color = c[n]
                marker = "o"
                s = 7**2
                label = f"{season} weekday"
            else:
                color = adjust_lightness(copy(c[n]), amount=0.8)
                marker = "D"
                s = 5.5**2
                label = f"{season} weekend"

            meter_season = meter_eval[
                (meter_eval["season"] == season) & (meter_eval["observed"].notna())
            ]
            meter_season = meter_season[meter_season["day_of_week"].isin(day_num)]

            T = meter_season["temperature"].values
            obs = meter_season["observed"].values
            model = meter_season["predicted"].values
            resid = obs - model

            ax[0].scatter(T, obs, color=color, marker=marker, s=s, label=label)
            if include_resid:
                ax[1].scatter(T, resid, color=color, marker=marker, s=s)

            if not plot_gaussian_ellipses:
                continue

            std_sqr = std = np.array(fit.model_settings.reduce_splits_num_std)[:, None]
            std_sqr = std.T * std

            mu, cov, a, b, phi = robust_confidence_ellipse(T, obs, std_sqr)

            ell = mpl.patches.Ellipse(
                mu, 2 * a, 2 * b, np.degrees(phi), color=color, zorder=10
            )
            ell.set_clip_box(ax[0].bbox)
            ell.set_alpha(0.3)
            ax[0].add_artist(ell)

    # Plot models
    for split in meter_eval["model_split"].unique():
        meter_segment = meter_eval[meter_eval["model_split"] == split]

        name = f"{split}__{meter_segment['model_type'].iloc[0]}"
        ax[0].plot(
            meter_segment["temperature"],
            meter_segment["predicted"],
            color="tab:orange",
            label=f"{name}",
        )

    # ax[0].plot(T, model["c_hdd_baseline"].model, color="tab:red", label=f"c_hdd_baseline")

    if include_resid:
        ax[1].axhline(y=0, linestyle=(0, (5, 1)), linewidth=1.5, color=(0.4, 0.4, 0.4))
        ax[0].get_shared_x_axes().join(ax[0], ax[1])
        ax[1].set_xlabel("Temperature", labelpad=10, fontsize=fontsize)
        ax[1].set_ylabel("Resid", labelpad=10, fontsize=fontsize)

    else:
        ax[0].set_xlabel("Temperature", labelpad=10, fontsize=fontsize)

    # ax.plot(hours, meter[:,2], linewidth=1.5, linestyle=(0, (6, 1)), color='firebrick')
    # ax.plot(hours, meter[:,2], linewidth=2.0, linestyle='-.')
    # ax.fill_between(hours, cg_lb, cg_ub, alpha=0.3, facecolor='peru')

    # ax.set_xlim([T[0], T[-1]])
    # ax.set_xticks(np.arange(0, 505, 168))
    ax[0].tick_params(axis="both", which="major", labelsize=0.85 * fontsize)

    if not plot_outliers:
        # Ignores crazy points when plotting based on iqr
        ylim = IQR_outlier(
            meter_eval["observed"].values, sigma_threshold=1.0, quantile=0.025
        )
        ylim_idx = [
            np.argmin(np.abs(x - meter_eval["observed"].values), axis=0) for x in ylim
        ]
        ylim = meter_eval["observed"].values[ylim_idx]
    else:
        ylim = np.quantile(meter_eval["observed"], [0, 1])

    ylim_border = 0.1 * (ylim[1] - ylim[0])
    ax[0].set_ylim([ylim[0] - ylim_border, ylim[1] + ylim_border])
    # ax.xaxis.set_minor_locator(mpl.ticker.AutoMinorLocator(7))
    # ax.tick_params(axis='both', which='major', labelsize=0.85*fontsize)
    # ax.yaxis.set_tick_params(which='minor', left=False)
    ax[0].set_ylabel("Usage", labelpad=10, fontsize=fontsize)

    legend = ax[0].legend(framealpha=0.0, fontsize=0.5 * fontsize)
    # legend._legend_box.align = 'left'

    plt.show()

    # if figsize is None:
    #     figsize = (10, 4)

    # if ax is None:
    #     fig, ax = plt.subplots(figsize=figsize)

    # color = "C1"
    # alpha = 1

    # temp_min, temp_max = (30, 90) if temp_range is None else temp_range

    # temps = np.arange(temp_min, temp_max)

    # prediction_index = pd.date_range(
    #     "2017-01-01T00:00:00Z", periods=len(temps), freq="D"
    # )

    # temps_daily = pd.Series(temps, index=prediction_index).resample("D").mean()
    # prediction = self._predict(temps_daily).model

    # plot_kwargs = {"color": color, "alpha": alpha or 0.3}
    # ax.plot(temps, prediction, **plot_kwargs)

    # if title is not None:
    #     ax.set_title(title)

    return ax



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
import numba
import numpy as np

from opendsm.common.adaptive_loss import adaptive_weights
from opendsm.common.utils import LN_MAX_POS_SYSTEM_VALUE, LN_MIN_POS_SYSTEM_VALUE


@numba.jit(nopython=True, error_model="numpy", cache=True)
def full_model(
    hdd_bp,
    hdd_beta,
    hdd_k,
    cdd_bp,
    cdd_beta,
    cdd_k,
    intercept,
    T_fit_bnds=np.array([]),
    T=np.array([]),
):
    """
    This function predicts the total energy consumption based on the given parameters.

    Parameters:
    hdd_bp (float): The base point for the heating model.
    hdd_beta (float): The beta value for the heating model.
    hdd_k (float): The k value for the heating model.
    cdd_bp (float): The base point for the cooling model.
    cdd_beta (float): The beta value for the cooling model.
    cdd_k (float): The k value for the cooling model.
    intercept (float): The intercept value for the model.
    T_fit_bnds (numpy array): The temperature bounds for the model fitting. Default is an empty numpy array.
    T (numpy array): The temperature values. Default is an empty numpy array.

    Returns:
    numpy array: The total energy consumption for each temperature value in T.
    """

    # if all variables are zero, return tidd model
    if (hdd_beta == 0) and (cdd_beta == 0):
        return np.ones_like(T) * intercept

    [T_min, T_max] = T_fit_bnds

    if cdd_bp < hdd_bp:
        hdd_bp, cdd_bp = cdd_bp, hdd_bp
        hdd_beta, cdd_beta = cdd_beta, hdd_beta
        hdd_k, cdd_k = cdd_k, hdd_k

    E_tot = np.empty_like(T)
    for n, Ti in enumerate(T):
        if (Ti < hdd_bp) or (
            (hdd_bp == cdd_bp) and (cdd_bp >= T_max)
        ):  # Temperature is within the heating model
            T_bp = hdd_bp
            beta = -hdd_beta
            k = hdd_k

        elif (Ti > cdd_bp) or (
            (hdd_bp == cdd_bp) and (hdd_bp <= T_min)
        ):  # Temperature is within the cooling model
            T_bp = cdd_bp
            beta = cdd_beta
            k = -cdd_k

        else:  # Temperature independent
            beta = 0.0

        # Evaluate
        if beta == 0:  # tidd
            E_tot[n] = intercept

        elif k == 0:  # c_hdd
            E_tot[n] = beta * (Ti - T_bp) + intercept

        else:  # smoothed c_hdd
            c_hdd = beta * (Ti - T_bp) + intercept

            exp_interior = 1 / k * (Ti - T_bp)
            exp_interior = np.clip(
                exp_interior, LN_MIN_POS_SYSTEM_VALUE, LN_MAX_POS_SYSTEM_VALUE
            )
            E_tot[n] = abs(beta * k) * (np.exp(exp_interior) - 1) + c_hdd

    return E_tot


@numba.jit(nopython=True, error_model="numpy", cache=True)
def get_full_model_x(model_key, x, T_min, T_max, T_min_seg, T_max_seg):
    """
    This function adjusts the parameters of a full model based on certain conditions.

    Parameters:
    x (list): A list containing the parameters of the model.
    T_min_seg (float): The minimum temperature segment.
    T_max_seg (float): The maximum temperature segment.

    Returns:
    list: A list of adjusted parameters.

    """

    if model_key == "hdd_tidd_cdd_smooth":
        [hdd_bp, hdd_beta, hdd_k, cdd_bp, cdd_beta, cdd_k, intercept] = x

    elif model_key == "hdd_tidd_cdd":
        [hdd_bp, hdd_beta, cdd_bp, cdd_beta, intercept] = x
        hdd_k = cdd_k = 0.0

    elif model_key == "c_hdd_tidd_smooth":
        [c_hdd_bp, c_hdd_beta, c_hdd_k, intercept] = x
        hdd_bp = cdd_bp = c_hdd_bp

        if c_hdd_beta < 0:
            hdd_beta = -c_hdd_beta
            hdd_k = c_hdd_k
            cdd_beta = cdd_k = 0.0

        else:
            cdd_beta = c_hdd_beta
            cdd_k = c_hdd_k
            hdd_beta = hdd_k = 0.0

    elif model_key == "c_hdd_tidd":
        [c_hdd_bp, c_hdd_beta, intercept] = x

        if c_hdd_bp < T_min_seg:
            cdd_bp = hdd_bp = T_min_seg
        elif c_hdd_bp > T_max_seg:
            cdd_bp = hdd_bp = T_max_seg
        else:
            hdd_bp = cdd_bp = c_hdd_bp

        if c_hdd_beta < 0:
            hdd_beta = -c_hdd_beta
            cdd_beta = cdd_k = hdd_k = 0.0

        else:
            cdd_beta = c_hdd_beta
            hdd_beta = hdd_k = cdd_k = 0.0

    elif model_key == "tidd":
        [intercept] = x
        hdd_bp = hdd_beta = hdd_k = cdd_bp = cdd_beta = cdd_k = 0.0

    x = [hdd_bp, hdd_beta, hdd_k, cdd_bp, cdd_beta, cdd_k, intercept]

    return fix_full_model_x(x, T_min, T_max)


@numba.jit(nopython=True, error_model="numpy", cache=True)
def fix_full_model_x(x, T_min_seg, T_max_seg):
    """
    This function adjusts the parameters of a full model based on certain conditions.

    Parameters:
    x (list): A list containing the parameters of the model [hdd_bp, hdd_beta, hdd_k, cdd_bp, cdd_beta, cdd_k, intercept].
    T_min_seg (float): The minimum temperature segment.
    T_max_seg (float): The maximum temperature segment.

    Returns:
    list: A list of adjusted parameters [hdd_bp, hdd_beta, hdd_k, cdd_bp, cdd_beta, cdd_k, intercept].

    """

    hdd_bp, hdd_beta, hdd_k, cdd_bp, cdd_beta, cdd_k, intercept = x

    # swap breakpoint order if they are reversed [hdd, cdd]
    if cdd_bp < hdd_bp:
        hdd_bp, cdd_bp = cdd_bp, hdd_bp
        hdd_beta, cdd_beta = cdd_beta, hdd_beta
        hdd_k, cdd_k = cdd_k, hdd_k

    # if there is a slope, but the breakpoint is at the end, it's a c_hdd_tidd model
    if hdd_bp != cdd_bp:
        if cdd_bp >= T_max_seg:
            cdd_beta = 0.0
        elif hdd_bp <= T_min_seg:
            hdd_beta = 0.0

    # if slopes are zero then smoothing is zero
    if hdd_beta == 0:
        hdd_k = 0.0

    if cdd_beta == 0:
        cdd_k = 0.0

    return [hdd_bp, hdd_beta, hdd_k, cdd_bp, cdd_beta, cdd_k, intercept]


def full_model_weight(
    hdd_bp,
    hdd_beta,
    hdd_k,
    cdd_bp,
    cdd_beta,
    cdd_k,
    intercept,
    T,
    residual,
    sigma=3.0,
    quantile=0.25,
    alpha=2.0,
    min_weight=0.0,
):
    """
    This function calculates the weights, C and alpha for a full model using adaptive weights

    Parameters:
    hdd_bp (float): The base point for heating degree days.
    hdd_beta (float): The beta value for heating degree days.
    hdd_k (float): The k value for heating degree days.
    cdd_bp (float): The base point for cooling degree days.
    cdd_beta (float): The beta value for cooling degree days.
    cdd_k (float): The k value for cooling degree days.
    intercept (float): The intercept of the model.
    T (array-like): The temperature array.
    residual (array-like): The residual array.
    weights (array-like): The input weights array.
    sigma (float, optional): The standard deviation. Default is 3.0.
    quantile (float, optional): The quantile to be used. Default is 0.25.
    alpha (float, optional): The alpha value. Default is 2.0.
    min_weight (float, optional): The minimum weight. Default is 0.0.

    Returns:
    tuple: Returns a tuple containing the weights, C and alpha for the full model.
    """

    if hdd_bp > cdd_bp:
        hdd_bp, cdd_bp = cdd_bp, hdd_bp

    if (hdd_beta == 0) and (cdd_beta == 0):  # intercept only
        resid_all = [residual]

    elif (cdd_bp >= T[-1]) or (hdd_bp <= T[0]):  # hdd or cdd only
        resid_all = [residual]

    elif hdd_beta == 0:
        idx_cdd_bp = np.argmin(np.abs(T - cdd_bp))

        resid_all = [residual[:idx_cdd_bp], residual[idx_cdd_bp:]]

    elif cdd_beta == 0:
        idx_hdd_bp = np.argmin(np.abs(T - hdd_bp))

        resid_all = [residual[:idx_hdd_bp], residual[idx_hdd_bp:]]

    else:
        idx_hdd_bp = np.argmin(np.abs(T - hdd_bp))
        idx_cdd_bp = np.argmin(np.abs(T - cdd_bp))

        if hdd_bp == cdd_bp:
            resid_all = [residual[:idx_hdd_bp], residual[idx_cdd_bp:]]

        else:
            resid_all = [
                residual[:idx_hdd_bp],
                residual[idx_hdd_bp:idx_cdd_bp],
                residual[idx_cdd_bp:],
            ]

    weight = []
    C = []
    a = []
    for resid in resid_all:
        if len(resid) == 0:
            continue

        elif len(resid) < 3:
            weight.append(np.ones_like(resid))
            C.append(np.ones_like(resid))
            a.append(np.ones_like(resid) * 2.0)

            continue

        _weight, _C, _a = adaptive_weights(
            resid, alpha=alpha, sigma=sigma, quantile=quantile, min_weight=min_weight
        )

        weight.append(_weight)
        C.append(np.ones_like(resid) * _C)
        a.append(np.ones_like(resid) * _a)

    weight_out = np.hstack(weight)
    C_out = np.hstack(weight)
    a_out = np.hstack(a)

    return weight_out, C_out, a_out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

from __future__ import annotations

import warnings

import numpy as np
import pandas as pd

from pydantic import BaseModel, ConfigDict

import pywt

from scipy.signal import find_peaks


from sklearn.preprocessing import StandardScaler, RobustScaler
from sklearn.decomposition import PCA, KernelPCA
from sklearn.cluster import DBSCAN, HDBSCAN, Birch, SpectralClustering



from opendsm.common.clustering import (
    bisect_k_means as _bisect_k_means,
    scoring as _scoring,
    settings as _settings,
)



class _LabelResult(BaseModel):
    """
    contains metrics about a cluster label returned from sklearn
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    labels: np.ndarray
    score: float
    score_unable_to_be_calculated: bool
    n_clusters: int


def _bisecting_kmeans_clustering(
    data: np.ndarray,
    settings: _settings.ClusteringSettings
):
    """
    clusters features using Bisecting K-Means algorithm
    """
    
    recluster_count = settings.bisecting_kmeans.recluster_count
    n_cluster_lower = settings.bisecting_kmeans.n_cluster.lower
    n_cluster_upper = settings.bisecting_kmeans.n_cluster.upper
    n_init = settings.bisecting_kmeans.internal_recluster_count
    inner_algorithm = settings.bisecting_kmeans.inner_algorithm
    bisecting_strategy = settings.bisecting_kmeans.bisecting_strategy

    score_choice = settings.bisecting_kmeans.scoring.score_metric
    dist_metric = settings.bisecting_kmeans.scoring.distance_metric
    min_cluster_size = settings.bisecting_kmeans.scoring.min_cluster_size
    max_non_outlier_cluster_count = 200

    seed = settings._seed

    results = []
    for i in range(recluster_count):
        algo = _bisect_k_means.BisectingKMeans(
            n_clusters=n_cluster_upper,
            init="k-means++",  # does not benefit from k-means++ like other k-means
            n_init=n_init,
            random_state=seed + i,
            algorithm=inner_algorithm,
            bisecting_strategy=bisecting_strategy,
        )
        algo.fit(data)
        labels_dict = algo.labels_full

        # if specifying clusters, only score the specified clusters
        if n_cluster_lower == n_cluster_upper:
            labels_dict = {n_cluster_lower: labels_dict[n_cluster_lower]}

        for n_cluster, labels in labels_dict.items():
            score, score_unable_to_be_calculated = _scoring.score_clusters(
                data,
                labels,
                n_cluster_lower,
                score_choice,
                dist_metric,
                min_cluster_size,
                max_non_outlier_cluster_count,
            )

            label_res = _LabelResult(
                labels=labels,
                score=score,
                score_unable_to_be_calculated=score_unable_to_be_calculated,
                n_clusters=n_cluster,
            )
            results.append(label_res)

    # get the results index with the smallest score
    HoF = None
    for result in results:
        if result.score_unable_to_be_calculated:
            continue

        if HoF is None or result.score < HoF.score:
            HoF = result

    return HoF.labels


def _birch_clustering(
    data: np.ndarray,
    settings: _settings.ClusteringSettings
):
    """
    Clusters features using Birch algorithm
    """

    n_cluster_lower = settings.birch.n_cluster.lower
    n_cluster_upper = settings.birch.n_cluster.upper
    threshold = settings.birch.threshold
    branching_factor = settings.birch.branching_factor

    score_choice = settings.birch.scoring.score_metric
    dist_metric = settings.birch.scoring.distance_metric
    min_cluster_size = settings.birch.scoring.min_cluster_size
    max_non_outlier_cluster_count = 200

    results = []
    for n_clusters in range(n_cluster_lower, n_cluster_upper + 1):
        algo = Birch(
            n_clusters=n_clusters,
            threshold=threshold,
            branching_factor=branching_factor,
        )
        labels = algo.fit_predict(data)
        
        # Calculate score for the clusters
        score, score_unable_to_be_calculated = _scoring.score_clusters(
                data,
                labels,
                n_cluster_lower,
                score_choice,
                dist_metric,
                min_cluster_size,
                max_non_outlier_cluster_count,
            )
        
        label_res = _LabelResult(
                labels=labels,
                score=score,
                score_unable_to_be_calculated=score_unable_to_be_calculated,
                n_clusters=n_clusters,
            )
        
        results.append(label_res)

    # get the results index with the smallest score
    HoF = None
    for result in results:
        if result.score_unable_to_be_calculated:
            continue

        if HoF is None or result.score < HoF.score:
            HoF = result

    return HoF.labels


def _dbscan_clustering(
    data: np.ndarray,
    settings: _settings.ClusteringSettings
):
    """
    clusters features using DBSCAN algorithm
    """
    algo = DBSCAN(
        eps=settings.dbscan.epsilon, 
        min_samples=settings.dbscan.min_samples, 
        metric=settings.dbscan.distance_metric.value,
        algorithm=settings.dbscan.nearest_neighbors_algorithm,
        leaf_size=settings.dbscan.leaf_size,
        p=settings.dbscan.minkowski_p,
    )
    labels = algo.fit_predict(data)

    return labels


def _hdbscan_clustering(
    data: np.ndarray,
    settings: _settings.ClusteringSettings
):
    """
    clusters features using HDBSCAN algorithm
    """
    min_samples = settings.hdbscan.min_samples
    if settings.hdbscan.min_samples == 1:
        min_samples = 2

    algo = HDBSCAN(
        min_samples=settings.hdbscan.scoring_sample_count, 
        min_cluster_size=min_samples,
        allow_single_cluster=settings.hdbscan.allow_single_cluster,
        max_cluster_size=settings.hdbscan.max_cluster_size,
        metric=settings.hdbscan.distance_metric,
        cluster_selection_epsilon=settings.hdbscan.cluster_selection_epsilon,
        alpha=settings.hdbscan.robust_single_linkage_scaling,
        algorithm=settings.hdbscan.nearest_neighbors_algorithm,
        leaf_size=settings.hdbscan.leaf_size,
        cluster_selection_method=settings.hdbscan.cluster_selection_method,
    )
    labels = algo.fit_predict(data)

    if settings.hdbscan.min_samples == 1:
        # get count of -1 labels
        outlier_count = np.sum(labels == -1)

        if outlier_count == 0:
            return labels

        # add to all labels to make room for outliers
        labels[labels != -1] += outlier_count

        # make labels with -1 defined as arange(max_label+1, n_samples)
        labels[labels == -1] = np.arange(0, outlier_count)

    return labels


def _spectral_clustering(
    data: np.ndarray,
    settings: _settings.ClusteringSettings
):
    """
    clusters features using Spectral Clustering algorithm
    """
    n_cluster_lower = settings.spectral.n_cluster.lower
    n_cluster_upper = settings.spectral.n_cluster.upper
    
    score_choice = settings.spectral.scoring.score_metric
    dist_metric = settings.spectral.scoring.distance_metric
    min_cluster_size = settings.spectral.scoring.min_cluster_size
    max_non_outlier_cluster_count = 200

    # transform data as spectral clustering doesn't like negative values
    # data = np.exp(-data / np.std(data))

    X = data
    results = []
    for n_clusters in range(n_cluster_lower, n_cluster_upper + 1):
        if n_clusters == n_cluster_lower:
            affinity = settings.spectral.affinity
        else:
            affinity = "precomputed"

        algo = SpectralClustering(
            n_clusters=n_clusters,
            eigen_solver=settings.spectral.eigen_solver,
            n_components=settings.spectral.n_components,
            affinity=affinity,
            n_neighbors=settings.spectral.nearest_neighbors,
            gamma=settings.spectral.gamma,
            eigen_tol=settings.spectral.eigen_tol,
            assign_labels=settings.spectral.assign_labels,
        )

        # hide UserWarning from sklearn
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=UserWarning)
            labels = algo.fit_predict(X)

        X = algo.affinity_matrix_

        # Calculate a score for the clustering
        score, score_unable_to_be_calculated = _scoring.score_clusters(
                data,
                labels,
                n_cluster_lower,
                score_choice,
                dist_metric,
                min_cluster_size,
                max_non_outlier_cluster_count,
            )
        
        label_res = _LabelResult(
                labels=labels,
                score=score,
                score_unable_to_be_calculated=score_unable_to_be_calculated,
                n_clusters=n_clusters,
            )
        
        results.append(label_res)

    # get the results index with the smallest score
    HoF = None
    for result in results:
        if result.score_unable_to_be_calculated:
            continue

        if HoF is None or result.score < HoF.score:
            HoF = result

    return HoF.labels


def _transform_data(
    data: np.ndarray,
    settings: _settings.ClusteringSettings
):
    """
    Transforms the data using the wavelet transform settings
    """
    settings = settings.transform_settings

    def _dwt_coeffs(data, wavelet="db1", wavelet_mode="periodization", n_levels=4):
        all_features = []
        # iterate through rows of numpy array
        for row in range(len(data)):
            # get max level of decomposition
            dwt_max_level = pywt.dwt_max_level(data[row].shape[0], wavelet)
            if n_levels > dwt_max_level:
                n_levels = dwt_max_level
            
            decomp_coeffs = pywt.wavedec(
                data[row], wavelet=wavelet, mode=wavelet_mode, level=n_levels
            )

            decomp_coeffs = np.hstack(decomp_coeffs)

            all_features.append(decomp_coeffs)

        return np.vstack(all_features)

    def _pca_coeffs(features, method, min_var_ratio_explained=0.95, n_components=None):
        if min_var_ratio_explained is not None:
            n_components = min_var_ratio_explained

        # kernel pca is not fully developed
        if method == "kernel_pca":
            if n_components ==  "mle":
                pca = PCA(n_components=n_components)
                pca_features = pca.fit_transform(features)

            pca = KernelPCA(n_components=None, kernel="rbf")
            pca_features = pca.fit_transform(features)

            if min_var_ratio_explained is not None:
                explained_variance_ratio = pca.eigenvalues_ / np.sum(pca.eigenvalues_)

                # get the cumulative explained variance ratio
                cumulative_explained_variance = np.cumsum(explained_variance_ratio)

                # find number of components that explain pct% of the variance
                n_components = np.argmax(cumulative_explained_variance > n_components).astype(int)

            if not isinstance(n_components, (int, np.integer)):
                raise ValueError("n_components must be an integer for kernel PCA")

            # pca = PCA(n_components=n_components)
            pca = KernelPCA(n_components=n_components, kernel="rbf")
            pca_features = pca.fit_transform(features)

        else:
            pca = PCA(n_components=n_components)
            pca_features = pca.fit_transform(features)

        return pca_features

    # calculate wavelet coefficients
    with warnings.catch_warnings():
        features = _dwt_coeffs(
            data, 
            settings.wavelet_name, 
            settings.wavelet_mode, 
            settings.wavelet_n_levels
        )

    pca_features = _pca_coeffs(
        features,
        settings.pca_method,
        settings.pca_min_variance_ratio_explained,
        settings.pca_n_components,
    )

    # normalize pca features
    if settings._standardize:
        pca_features = (pca_features - pca_features.mean()) / pca_features.std()

    if settings.pca_include_median:
        pca_features = np.hstack([pca_features, np.median(data, axis=1)[:, None]])

    return pca_features


def cluster_reorder(
    data: pd.DataFrame, 
    cluster_labels: np.ndarray,
    settings: _settings.ClusteringSettings,
):
    sort_method = settings.cluster_sort_options.method
    agg_type = settings.cluster_sort_options.aggregation
    reverse = settings.cluster_sort_options.reverse

    # assign labels to data
    df = data.copy()
    df["label"] = cluster_labels
    
    # group by cluster and aggregate
    df_cluster = df.groupby('label').agg(agg_type)
    n_clusters = len(df_cluster)

    if sort_method == "size":
        # sort clusters by count
        cluster_size = df['label'].value_counts()
        cluster_size = cluster_size.sort_values()

        features = cluster_size

    elif sort_method == "peak":
        # TODO: This is a work in progress

        # subtract each cluster's median from the cluster's median
        df_cluster_norm = df_cluster.sub(df_cluster.agg(agg_type, axis=1), axis=0)
        cluster_max = df_cluster_norm.abs().max().max()
        df_cluster_norm = df_cluster_norm/cluster_max

        # define threshold for peak and valley
        threshold = np.quantile(abs(df_cluster.values), 0.75)

        # find peaks and valleys
        peak = {}
        valley = {}
        norm = {}
        for i in range(n_clusters):
            cluster_normal = df_cluster.iloc[i]
            norm[i] = cluster_normal.agg(agg_type)
            df_cluster_norm = cluster_normal - norm[i]
            thresh = threshold - norm[i]

            peak[i] = find_peaks(df_cluster_norm.values, height=thresh, width=1)[0]
            valley[i] = find_peaks(-df_cluster_norm.values, height=thresh, width=1)[0]

            if len(peak[i]) == 0:
                peak[i] = None
            else:
                peak[i] = peak[i][0]

            if len(valley[i]) == 0:
                valley[i] = None
            else:
                valley[i] = valley[i][0]
        
        # create df with peak and valley
        features = pd.DataFrame({'peak': peak, 'valley': valley, "norm": norm})

        features = features.sort_values(by=["peak", "valley", "norm"], na_position='first')

    # create dictionary to remap cluster numbers to features order
    if not reverse:
        cluster_map = {features.index[i]: i for i in range(n_clusters)}
    else:
        cluster_map = {features.index[i]: i for i in range(n_clusters)[::-1]}

    return cluster_map



def cluster_features(
    df: pd.DataFrame,
    settings: _settings.ClusteringSettings,
):
    # convert data to numpy array
    data = df.to_numpy()
    
    # bypass clustering if cluster count is >= data
    if settings.algorithm_selection not in ["dbscan", "hdbscan"]:
        algo = getattr(settings, f"{settings.algorithm_selection.value}")
        if algo.n_cluster.lower >= len(data):
            return np.arange(len(data))

    # standardize the data
    if settings.standardize:
        data = (data - data.mean()) / data.std()

    # transform the data
    if settings.transform_settings is not None:
        data = _transform_data(data, settings)

    # cluster the pca features
    if settings.algorithm_selection == "bisecting_kmeans":
        cluster_fcn = _bisecting_kmeans_clustering
    elif settings.algorithm_selection == "birch":
        cluster_fcn = _birch_clustering
    elif settings.algorithm_selection == "dbscan":
        cluster_fcn = _dbscan_clustering
    elif settings.algorithm_selection == "hdbscan":
        cluster_fcn = _hdbscan_clustering
    elif settings.algorithm_selection == "spectral":
        cluster_fcn = _spectral_clustering
    else:
        raise ValueError(f"Unknown clustering algorithm: {settings.algorithm_selection}")
    
    cluster_labels = cluster_fcn(data, settings)

    if settings.sort_clusters:
        cluster_remap_dict = cluster_reorder(df, cluster_labels, settings)

        # remap cluster labels using cluster_remap_dict
        cluster_labels = np.vectorize(cluster_remap_dict.get)(cluster_labels)

    return cluster_labels


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import numpy as np
import pandas as pd

import pydantic
from typing import Optional
from enum import Enum

from functools import cached_property  # TODO: This requires Python 3.8

from opendsm.common.utils import median_absolute_deviation, t_stat
from opendsm.common.pydantic_utils import (
    ArbitraryPydanticModel,
    PydanticDf,
    PydanticFromDict,
)


def computed_field_cached_property():
    decs = [pydantic.computed_field, cached_property]

    def deco(f):
        for dec in reversed(decs):
            f = dec(f)
        return f

    return deco


class ColumnMetrics(ArbitraryPydanticModel):
    series: pd.Series = pydantic.Field(
        exclude=True,
        repr=False,
    )

    @computed_field_cached_property()
    def sum(self) -> float:
        return self.series.sum()

    @computed_field_cached_property()
    def mean(self) -> float:
        return self.sum / len(self.series)

    @computed_field_cached_property()
    def variance(self) -> float:
        return self.series.var(ddof=0)

    @computed_field_cached_property()
    def std(self) -> float:
        return self.variance**0.5

    @computed_field_cached_property()
    def cvstd(self) -> float:
        return self.std / self.mean

    @computed_field_cached_property()
    def sum_squared(self) -> float:
        return (self.series**2).sum()

    @computed_field_cached_property()
    def median(self) -> float:
        return self.series.median()

    @computed_field_cached_property()
    def MAD_scaled(self) -> float:
        return median_absolute_deviation(self.series, self.median)

    @computed_field_cached_property()
    def iqr(self) -> float:
        return np.diff(np.quantile(self.series, [0.25, 0.75]))[0]

    @computed_field_cached_property()
    def skew(self) -> float:
        return self.series.skew()

    @computed_field_cached_property()
    def kurtosis(self) -> float:
        return self.series.kurtosis()


def _safe_divide(numerator, denominator, min_denominator=1e-3):
    if denominator <= min_denominator and numerator > 10 * min_denominator:
        return None

    return numerator / denominator


class BaselineMetrics(ArbitraryPydanticModel):
    # TODO: Update the doc string
    """Contains measures of model fit and summary statistics on the input dataframe.

    Parameters
    ----------
    observed_input : :any:`pandas.Series`
        Series with :any:`pandas.DatetimeIndex` with a set of electricity or
        gas meter values.
    predicted_input : :any:`pandas.Series`
        Series with :any:`pandas.DatetimeIndex` with a set of electricity or
        gas meter values.
    num_parameters : :any:`int`, optional
        The number of parameters (excluding the intercept) used in the
        regression from which the predictions were derived.
    autocorr_lags : :any:`int`, optional
        The number of lags to use when calculating the autocorrelation of the
        residuals.
    confidence_level : :any:`int`, optional
        Confidence level used in fractional savings uncertainty computations.

    Attributes
    ----------
    observed_length : :any:`int`
        The length of the observed.
    predicted_length : :any:`int`
        The length of the predicted.
    merged_length : :any:`int`
        The length of the dataframe resulting from the inner join of the
        observed and the predicted.
    observed_mean : :any:`float`
        The mean of the observed.
    predicted_mean : :any:`float`
        The mean of the predicted.
    observed_skew : :any:`float`
        The skew of the observed.
    predicted_skew : :any:`float`
        The skew of the predicted.
    observed_kurtosis : :any:`float`
        The excess kurtosis of the observed.
    predicted_kurtosis : :any:`float`
        The excess kurtosis of the predicted.
    observed_cvstd : :any:`float`
        The coefficient of standard deviation of the observed.
    predicted_cvstd : :any:`float`
        The coefficient of standard deviation of the predicted.
    r_squared : :any:`float`
        The r-squared of the model from which the predicted was produced.
    r_squared_adj : :any:`float`
        The r-squared of the predicted relative to the
        observed, adjusted by the number of parameters in the model.
    cvrmse : :any:`float`
        The coefficient of variation (root-mean-squared error) of the
        predicted relative to the observed.
    cvrmse_adj : :any:`float`
        The coefficient of variation (root-mean-squared error) of the
        predicted relative to the observed, adjusted
        by the number of parameters in the model.
    mape : :any:`float`
        The mean absolute percent error of the predicted relative
        to the observed.
    mape_no_zeros : :any:`float`
        The mean absolute percent error of the predicted relative
        to the observed, with all time periods dropped where the
        observed was not greater than zero.
    num_meter_zeros : :any:`int`
        The number of time periods for which the observed was not
        greater than zero.
    nmae : :any:`float`
        The normalized mean absolute error of the predicted
        relative to the observed.
    nmbe : :any:`float`
        The normalized mean bias error of the predicted relative
        to the observed.
    autocorr_resid : :any:`float`
        The autocorrelation of the residuals (where the residuals equal the
        predicted minus the observed), measured
        using a number of lags equal to autocorr_lags.
    n_prime: :any:`float`
        The number of baseline inputs corrected for autocorrelation -- used
        in fractional savings uncertainty computation.
    single_tailed_confidence_level: :any:`float`
        The adjusted confidence level for use in single-sided tests.
    degrees_of_freedom: :any:`float
        Maxmimum number of independent variables which have the freedom to vary
    t_stat: :any:`float
        t-statistic, used for hypothesis testing
    cvrmse_auto_corr_correction: :any:`float
        Correctoin factor the apply to cvrmse to account for autocorrelation of inputs.
    approx_factor_auto_corr_correction: :any:`float
        Approximation factor used in ashrae 14 guideline for uncertainty computation.
    fsu_base_term: :any:`float
        Base term used in fractional savings uncertainty computation.

    """

    """Input dataframe to be used for metrics calculations"""
    df: pd.DataFrame = pydantic.Field(
        exclude=True,
        repr=False,
    )

    _min_denominator: float = 1e-3

    """Number of model parameters"""
    num_model_params: int = pydantic.Field(
        ge=1,
        validate_default=True,
    )

    @cached_property
    def _df(self) -> pd.DataFrame:
        _df = self.df[["observed", "predicted"]].copy()

        if len(_df) < 1:
            raise ValueError("Input dataframe must have at least one row")

        # Check dataframe
        expected_columns = {"observed": "float", "predicted": "float"}
        _df = PydanticDf(df=_df, column_types=expected_columns).df

        # drop non finite values from df
        _df = _df[np.isfinite(_df["observed"]) & np.isfinite(_df["predicted"])]

        # get residuals
        _df["residuals"] = _df["observed"] - _df["predicted"]

        return _df

    @computed_field_cached_property()
    def n(self) -> float:
        return len(self._df)

    @computed_field_cached_property()
    def n_prime(self) -> float:
        # lag should be 1 according to https://www.osti.gov/servlets/purl/1366449
        autocorr = self._df["residuals"].autocorr(lag=1)

        _n_prime = float(self.n * (1 - autocorr) / (1 + autocorr))

        if not np.isfinite(_n_prime):
            # TODO: Create warning
            _n_prime = 1

        return _n_prime

    @computed_field_cached_property()
    def ddof(self) -> float:
        _ddof = self.n - self.num_model_params

        if _ddof < 1:
            # TODO: Create warning
            _ddof = 1

        return _ddof

    @computed_field_cached_property()
    def ddof_autocorr(self) -> float:
        # TODO: should this be rounded?
        _ddof_autocorr = self.n_prime - self.num_model_params

        # TODO: what to do if less than 1?
        if _ddof_autocorr < 1:
            # TODO: Create warning
            _ddof_autocorr = 1

        return _ddof_autocorr

    @computed_field_cached_property()
    def observed(self) -> ColumnMetrics:
        return ColumnMetrics(series=self._df["observed"])

    @computed_field_cached_property()
    def predicted(self) -> ColumnMetrics:
        return ColumnMetrics(series=self._df["predicted"])

    @computed_field_cached_property()
    def residuals(self) -> ColumnMetrics:
        return ColumnMetrics(series=self._df["residuals"])

    @computed_field_cached_property()
    def mae(self) -> float:
        return self._df["residuals"].abs().mean()

    @computed_field_cached_property()
    def nmae(self) -> Optional[float]:
        return _safe_divide(self.mae, self.observed.mean, self._min_denominator)

    @computed_field_cached_property()
    def pnmae(self) -> Optional[float]:
        return _safe_divide(self.mae, self.observed.iqr, self._min_denominator)

    @computed_field_cached_property()
    def mbe(self) -> float:
        return self.residuals.mean

    @computed_field_cached_property()
    def nmbe(self) -> Optional[float]:
        return _safe_divide(self.mbe, self.observed.mean, self._min_denominator)

    @computed_field_cached_property()
    def pnmbe(self) -> Optional[float]:
        return _safe_divide(self.mbe, self.observed.iqr, self._min_denominator)

    @computed_field_cached_property()
    def sse(self) -> float:
        return self.residuals.sum_squared

    @computed_field_cached_property()
    def mse(self) -> float:
        return self.sse / self.n

    @computed_field_cached_property()
    def rmse(self) -> float:
        return self.mse**0.5

    @computed_field_cached_property()
    def rmse_adj(self) -> float:
        return (self.sse / self.ddof) ** 0.5

    @computed_field_cached_property()
    def rmse_autocorr_adj(self) -> float:
        return (self.sse / self.ddof_autocorr) ** 0.5

    @computed_field_cached_property()
    def cvrmse(self) -> Optional[float]:
        return _safe_divide(self.rmse, self.observed.mean, self._min_denominator)

    @computed_field_cached_property()
    def cvrmse_adj(self) -> Optional[float]:
        return _safe_divide(self.rmse_adj, self.observed.mean, self._min_denominator)

    @computed_field_cached_property()
    def cvrmse_autocorr_adj(self) -> Optional[float]:
        return _safe_divide(
            self.rmse_autocorr_adj, self.observed.mean, self._min_denominator
        )

    @computed_field_cached_property()
    def pnrmse(self) -> Optional[float]:
        return _safe_divide(self.rmse, self.observed.iqr, self._min_denominator)

    @computed_field_cached_property()
    def pnrmse_adj(self) -> Optional[float]:
        return _safe_divide(self.rmse_adj, self.observed.iqr, self._min_denominator)

    @computed_field_cached_property()
    def pnrmse_autocorr_adj(self) -> Optional[float]:
        return _safe_divide(
            self.rmse_autocorr_adj, self.observed.iqr, self._min_denominator
        )

    @computed_field_cached_property()
    def r_squared(self) -> float:
        return self._df[["predicted", "observed"]].corr().iloc[0, 1] ** 2

    @computed_field_cached_property()
    def r_squared_adj(self) -> Optional[float]:
        n = self.n
        n_adj = self.ddof

        num = (1 - self.r_squared) * (n - 1)
        den = n_adj - 1

        res = _safe_divide(num, den, self._min_denominator)
        if res is None:
            return None
        return 1 - res

    @computed_field_cached_property()
    def mape(self) -> Optional[float]:
        df = self._df
        df_no_zeros = df[np.abs(df["observed"]) >= self._min_denominator]

        if len(df_no_zeros) == 0:
            return None

        return (df_no_zeros["residuals"] / df_no_zeros["observed"]).abs().mean()


def BaselineMetricsFromDict(input_dict):
    for k in ["observed", "predicted", "residuals"]:
        input_dict[k] = PydanticFromDict(input_dict[k], name="ColumnMetrics")

    return PydanticFromDict(input_dict, name="BaselineMetrics")


class ModelChoice(str, Enum):
    HOURLY = "hourly"
    HOURLYSOLAR = "hourly"
    DAILY = "daily"
    BILLING = "billing"


class ReportingMetrics(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    baseline_metrics: BaselineMetrics = pydantic.Field(exclude=True)

    """Reporting dataframe to be used for metrics calculations"""
    reporting_df: pd.DataFrame = pydantic.Field(exclude=True)

    """Data frequency of the model for use in uncertainty calculations"""
    data_frequency: ModelChoice = pydantic.Field(exclude=False)

    """Confidence level used in uncertainty calculations"""
    confidence_level: float = pydantic.Field(
        ge=0.0,
        le=1.0,
        default=0.90,
        validate_default=True,
    )

    """Number of tails to use in uncertainty calculations"""
    t_tail: int = pydantic.Field(
        ge=1,
        le=2,
        default=2,  # ASHRAE 14 uses 2 tail
        validate_default=True,
    )

    @property
    def _baseline(self) -> BaselineMetrics:
        return self.baseline_metrics

    @cached_property
    def _df(self) -> pd.DataFrame:
        _df = self.reporting_df[["observed", "predicted"]].copy()

        if len(_df) < 1:
            raise ValueError("Input dataframe must have at least one row")

        # Check dataframe
        expected_columns = {"observed": "float", "predicted": "float"}
        _df = PydanticDf(df=_df, column_types=expected_columns).df

        # drop non finite values from df
        _df = _df[np.isfinite(_df["observed"]) & np.isfinite(_df["predicted"])]

        # # get residuals
        # _df["residuals"] = _df["observed"] - _df["predicted"]

        return _df

    @computed_field_cached_property()
    def n(self) -> float:
        return len(self._df)

    @computed_field_cached_property()
    def observed_sum(self) -> float:
        return self._df["observed"].sum()

    @computed_field_cached_property()
    def predicted_sum(self) -> float:
        return self._df["predicted"].sum()

    @computed_field_cached_property()
    def t_stat(self) -> float:
        return t_stat(1 - self.confidence_level, self._baseline.ddof, tail=self.t_tail)

    @computed_field_cached_property()
    def savings(self) -> float:
        return self.predicted_sum - self.observed_sum

    @computed_field_cached_property()
    def total_savings_uncertainty(self) -> float:
        E_reporting = self.predicted_sum
        n = self._baseline.n
        n_prime = self._baseline.n_prime
        m = self.n
        t = self.t_stat
        cvrmse_autocorr_adj = self._baseline.cvrmse_autocorr_adj

        # part of approximation factor used in ashrae 14 guideline
        approx_factor = np.sqrt(n / (m * n_prime) * (1 + (2 / n_prime)))

        s_unc_base = E_reporting * (t * cvrmse_autocorr_adj * approx_factor)

        if self.data_frequency == "hourly":
            s_unc = 1.26 * s_unc_base

        elif self.data_frequency in ["daily", "billing"]:
            M = len(self._df.index.month.unique())

            if self.data_frequency == "daily":
                coefs = [-0.00024, 0.03535, 1.00286]
            else:
                coefs = [-0.00022, 0.03306, 0.94054]

            s_unc = np.polyval(coefs, M) * s_unc_base

        else:
            raise ValueError("model_type must be 'hourly', 'daily', or 'billing'")

        return s_unc

    @computed_field_cached_property()
    def fsu(self) -> float:
        return self.total_savings_uncertainty / self.savings

    @computed_field_cached_property()
    def predicted_data_point_unc(self) -> float:
        return self.total_savings_uncertainty / np.sqrt(self.n)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
import gzip
from tempfile import TemporaryFile
import importlib.resources
import platform

import pandas as pd
import pytest

from opendsm.eemeter.utilities.io import (
    meter_data_from_csv,
    meter_data_from_json,
    meter_data_to_csv,
    temperature_data_from_csv,
    temperature_data_from_json,
    temperature_data_to_csv,
)


def test_meter_data_from_csv(sample_metadata):
    meter_item = sample_metadata["il-electricity-cdd-hdd-daily"]
    meter_data_filename = meter_item["meter_data_filename"]

    fname = str(
        importlib.resources.files("opendsm.eemeter.samples").joinpath(
            meter_data_filename
        )
    )
    with gzip.open(fname) as f:
        meter_data = meter_data_from_csv(f)
    assert meter_data.shape == (810, 1)
    assert str(meter_data.index.tz) == "UTC"
    assert meter_data.index.freq is None


def test_meter_data_from_csv_gzipped(sample_metadata):
    meter_item = sample_metadata["il-electricity-cdd-hdd-daily"]
    meter_data_filename = meter_item["meter_data_filename"]

    with importlib.resources.files("opendsm.eemeter.samples").joinpath(
        meter_data_filename
    ).open("rb") as f:
        meter_data = meter_data_from_csv(f, gzipped=True)
    assert meter_data.shape == (810, 1)
    assert str(meter_data.index.tz) == "UTC"
    assert meter_data.index.freq is None


def test_meter_data_from_csv_with_tz(sample_metadata):
    meter_item = sample_metadata["il-electricity-cdd-hdd-daily"]
    meter_data_filename = meter_item["meter_data_filename"]

    with importlib.resources.files("opendsm.eemeter.samples").joinpath(
        meter_data_filename
    ).open("rb") as f:
        meter_data = meter_data_from_csv(f, gzipped=True, tz="US/Eastern")
    assert meter_data.shape == (810, 1)
    assert str(meter_data.index.tz) == "US/Eastern"
    assert meter_data.index.freq is None


def test_meter_data_from_csv_hourly_freq(sample_metadata):
    meter_item = sample_metadata["il-electricity-cdd-hdd-daily"]
    meter_data_filename = meter_item["meter_data_filename"]

    with importlib.resources.files("opendsm.eemeter.samples").joinpath(
        meter_data_filename
    ).open("rb") as f:
        meter_data = meter_data_from_csv(f, gzipped=True, freq="hourly")
    assert meter_data.shape == (19417, 1)
    assert str(meter_data.index.tz) == "UTC"
    assert meter_data.index.freq == "h"


def test_meter_data_from_csv_daily_freq(sample_metadata):
    meter_item = sample_metadata["il-electricity-cdd-hdd-daily"]
    meter_data_filename = meter_item["meter_data_filename"]

    with importlib.resources.files("opendsm.eemeter.samples").joinpath(
        meter_data_filename
    ).open("rb") as f:
        meter_data = meter_data_from_csv(f, gzipped=True, freq="daily")
    assert meter_data.shape == (810, 1)
    assert str(meter_data.index.tz) == "UTC"
    assert meter_data.index.freq == "D"


def test_meter_data_from_csv_custom_columns(sample_metadata):
    with TemporaryFile() as f:
        f.write(b"start_custom,kWh\n" b"2017-01-01T00:00:00,10\n")
        f.seek(0)
        meter_data = meter_data_from_csv(f, start_col="start_custom", value_col="kWh")
    assert meter_data.shape == (1, 1)
    assert str(meter_data.index.tz) == "UTC"
    assert meter_data.index.freq is None


def test_meter_data_from_json_none(sample_metadata):
    data = None
    meter_data = meter_data_from_json(data)
    assert meter_data.shape == (0, 1)
    assert str(meter_data.index.tz) == "UTC"
    assert meter_data.index.freq is None


def test_meter_data_from_json_orient_list(sample_metadata):
    data = [["2017-01-01T00:00:00Z", 11], ["2017-01-02T00:00:00Z", 10]]
    meter_data = meter_data_from_json(data, orient="list")
    assert meter_data.shape == (2, 1)
    assert str(meter_data.index.tz) == "UTC"
    assert meter_data.index.freq is None


def test_meter_data_from_json_orient_list_empty(sample_metadata):
    data = []
    meter_data = meter_data_from_json(data)
    assert meter_data.shape == (0, 1)
    assert str(meter_data.index.tz) == "UTC"
    assert meter_data.index.freq is None


def test_meter_data_from_json_orient_records(sample_metadata):
    data = [
        {"start": "2017-01-01T00:00:00Z", "value": 11},
        {"start": "2017-01-02T00:00:00Z", "value": ""},
        {"start": "2017-01-03T00:00:00Z", "value": 10},
    ]
    meter_data = meter_data_from_json(data, orient="records")
    assert meter_data.shape == (3, 1)
    assert str(meter_data.index.tz) == "UTC"
    assert meter_data.index.freq is None


def test_meter_data_from_json_orient_records_empty(sample_metadata):
    data = []
    meter_data = meter_data_from_json(data, orient="records")
    assert meter_data.shape == (0, 1)
    assert str(meter_data.index.tz) == "UTC"
    assert meter_data.index.freq is None


def test_meter_data_from_json_orient_records_with_estimated_true(sample_metadata):
    data = [
        {"start": "2017-01-01T00:00:00Z", "value": 11, "estimated": True},
        {"start": "2017-01-02T00:00:00Z", "value": 10, "estimated": "true"},
        {"start": "2017-01-03T00:00:00Z", "value": 10, "estimated": "True"},
        {"start": "2017-01-04T00:00:00Z", "value": 10, "estimated": "1"},
        {"start": "2017-01-05T00:00:00Z", "value": 10, "estimated": 1},
    ]
    meter_data = meter_data_from_json(data, orient="records")
    assert meter_data.shape == (5, 2)
    assert str(meter_data.index.tz) == "UTC"
    assert meter_data.index.freq is None
    assert meter_data.estimated.sum() == 5


def test_meter_data_from_json_orient_records_with_estimated_false(sample_metadata):
    data = [
        {"start": "2017-01-01T00:00:00Z", "value": 10, "estimated": False},
        {"start": "2017-01-02T00:00:00Z", "value": 10, "estimated": "false"},
        {"start": "2017-01-03T00:00:00Z", "value": 10, "estimated": "False"},
        {"start": "2017-01-04T00:00:00Z", "value": 10, "estimated": ""},
        {"start": "2017-01-05T00:00:00Z", "value": 10, "estimated": None},
        {"start": "2017-01-05T00:00:00Z", "value": 10},
    ]
    meter_data = meter_data_from_json(data, orient="records")
    assert meter_data.shape == (6, 2)
    assert str(meter_data.index.tz) == "UTC"
    assert meter_data.index.freq is None
    assert meter_data.estimated.sum() == 0


def test_meter_data_from_json_bad_orient(sample_metadata):
    data = [["2017-01-01T00:00:00Z", 11], ["2017-01-02T00:00:00Z", 10]]
    with pytest.raises(ValueError):
        meter_data_from_json(data, orient="NOT_ALLOWED")


def test_meter_data_to_csv(sample_metadata):
    df = pd.DataFrame(
        {"value": [5]}, index=pd.to_datetime(["2017-01-01T00:00:00Z"], utc=True)
    )
    with TemporaryFile("w+") as f:
        meter_data_to_csv(df, f)
        f.seek(0)
        if platform.system() == "Windows":
            assert f.read() == ("start,value\n\n" "2017-01-01 00:00:00+00:00,5\n\n")
        else:
            assert f.read() == ("start,value\n" "2017-01-01 00:00:00+00:00,5\n")


def test_temperature_data_from_csv(sample_metadata):
    meter_item = sample_metadata["il-electricity-cdd-hdd-daily"]
    temperature_filename = meter_item["temperature_filename"]

    fname = str(
        importlib.resources.files("opendsm.eemeter.samples").joinpath(
            temperature_filename
        )
    )
    with gzip.open(fname) as f:
        temperature_data = temperature_data_from_csv(f)
    assert temperature_data.shape == (19417,)
    assert str(temperature_data.index.tz) == "UTC"
    assert temperature_data.index.freq is None


def test_temperature_data_from_csv_gzipped(sample_metadata):
    meter_item = sample_metadata["il-electricity-cdd-hdd-daily"]
    temperature_filename = meter_item["temperature_filename"]

    with importlib.resources.files("opendsm.eemeter.samples").joinpath(
        temperature_filename
    ).open("rb") as f:
        temperature_data = temperature_data_from_csv(f, gzipped=True)
    assert temperature_data.shape == (19417,)
    assert str(temperature_data.index.tz) == "UTC"
    assert temperature_data.index.freq is None


def test_temperature_data_from_csv_with_tz(sample_metadata):
    meter_item = sample_metadata["il-electricity-cdd-hdd-daily"]
    temperature_filename = meter_item["temperature_filename"]

    with importlib.resources.files("opendsm.eemeter.samples").joinpath(
        temperature_filename
    ).open("rb") as f:
        temperature_data = temperature_data_from_csv(f, gzipped=True, tz="US/Eastern")
    assert temperature_data.shape == (19417,)
    assert str(temperature_data.index.tz) == "US/Eastern"
    assert temperature_data.index.freq is None


def test_temperature_data_from_csv_hourly_freq(sample_metadata):
    meter_item = sample_metadata["il-electricity-cdd-hdd-daily"]
    temperature_filename = meter_item["temperature_filename"]

    with importlib.resources.files("opendsm.eemeter.samples").joinpath(
        temperature_filename
    ).open("rb") as f:
        temperature_data = temperature_data_from_csv(f, gzipped=True, freq="hourly")
    assert temperature_data.shape == (19417,)
    assert str(temperature_data.index.tz) == "UTC"
    assert temperature_data.index.freq == "h"


def test_temperature_data_from_csv_custom_columns(sample_metadata):
    with TemporaryFile() as f:
        f.write(b"dt_custom,tempC\n" b"2017-01-01T00:00:00,10\n")
        f.seek(0)
        temperature_data = temperature_data_from_csv(
            f, date_col="dt_custom", temp_col="tempC"
        )
    assert temperature_data.shape == (1,)
    assert str(temperature_data.index.tz) == "UTC"
    assert temperature_data.index.freq is None


def test_temperature_data_from_json_orient_list(sample_metadata):
    data = [["2017-01-01T00:00:00Z", 11], ["2017-01-02T00:00:00Z", 10]]
    temperature_data = temperature_data_from_json(data, orient="list")
    assert temperature_data.shape == (2,)
    assert str(temperature_data.index.tz) == "UTC"
    assert temperature_data.index.freq is None


def test_temperature_data_from_json_bad_orient(sample_metadata):
    data = [["2017-01-01T00:00:00Z", 11], ["2017-01-02T00:00:00Z", 10]]
    with pytest.raises(ValueError):
        temperature_data_from_json(data, orient="NOT_ALLOWED")


def test_temperature_data_to_csv(sample_metadata):
    series = pd.Series(10, index=pd.to_datetime(["2017-01-01T00:00:00Z"], utc=True))
    with TemporaryFile("w+") as f:
        temperature_data_to_csv(series, f)
        f.seek(0)
        if platform.system() == "Windows":
            assert f.read() == ("dt,temperature\n\n" "2017-01-01 00:00:00+00:00,10\n\n")
        else:
            assert f.read() == ("dt,temperature\n" "2017-01-01 00:00:00+00:00,10\n")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import numpy as np
import pytest
from scipy.stats import linregress, theilslopes

from opendsm.eemeter.models.daily.utilities.settings import DailySettings
from opendsm.eemeter.models.daily.utilities.base_model import (
    get_slope,
    linear_fit,
    get_smooth_coeffs,
    fix_identical_bnds,
    get_intercept,
)


@pytest.fixture
def get_settings():
    settings = DailySettings()
    return settings


def test_get_intercept():
    # Test case 1: alpha = 2, y has positive values
    y = np.array([1, 2, 3, 4, 5])
    assert get_intercept(y) == 3.0

    # Test case 2: alpha = 2, y has negative values
    y = np.array([-5, -4, -3, -2, -1])
    assert get_intercept(y) == -3.0

    # Test case 3: alpha = 1, y has positive values
    y = np.array([1, 2, 3, 4, 5])
    assert get_intercept(y, alpha=1) == 3.0

    # Test case 4: alpha = 1, y has negative values
    y = np.array([-5, -4, -3, -2, -1])
    assert get_intercept(y, alpha=1) == -3.0

    # Test case 5: alpha = 2, y has both positive and negative values
    y = np.array([-5, -4, -3, 2, 1])
    assert get_intercept(y) == -1.8

    # Test case 6: alpha = 1, y has both positive and negative values
    y = np.array([-5, -4, -3, 2, 1])
    assert get_intercept(y, alpha=1) == -3.0


def test_get_slope():
    # Test case 1: Test with alpha=2
    x = np.array([1, 2, 3, 4, 5])
    y = np.array([2, 4, 6, 8, 10])
    x_bp = 3
    intercept = 0
    alpha = 2
    expected_slope = 2.0102
    assert np.isclose(
        get_slope(x, y, x_bp, intercept, alpha), expected_slope, atol=1e-3
    )

    # Test case 2: Test with alpha=1
    x = np.array([1, 2, 3, 4, 5])
    y = np.array([2, 4, 6, 8, 10])
    x_bp = 3
    intercept = 0
    alpha = 1
    expected_slope = 2.125
    assert np.isclose(
        get_slope(x, y, x_bp, intercept, alpha), expected_slope, atol=1e-3
    )

    # Test case 3: Test with negative y values
    x = np.array([1, 2, 3, 4, 5])
    y = np.array([-2, -4, -6, -8, -10])
    x_bp = 3
    intercept = 0
    alpha = 2
    expected_slope = -2.016
    assert np.isclose(
        get_slope(x, y, x_bp, intercept, alpha), expected_slope, atol=1e-3
    )

    # Test case 4: Test with non-zero intercept
    x = np.array([1, 2, 3, 4, 5])
    y = np.array([2, 4, 6, 8, 10])
    x_bp = 3
    intercept = 1
    alpha = 2
    expected_slope = 2.0143
    assert np.isclose(
        get_slope(x, y, x_bp, intercept, alpha), expected_slope, atol=1e-3
    )


def test_linear_fit():
    # Test case 1: Test with alpha = 2
    x = np.array([1, 2, 3, 4, 5])
    y = np.array([2, 4, 6, 8, 10])
    alpha = 2
    slope, intercept = linear_fit(x, y, alpha)
    res = linregress(x, y)
    assert slope == res.slope
    assert intercept == res.intercept

    # Test case 2: Test with alpha = 0.95
    x = np.array([1, 2, 3, 4, 5])
    y = np.array([2, 4, 6, 8, 10])
    alpha = 0.95
    slope, intercept = linear_fit(x, y, alpha)
    res = theilslopes(y, x, alpha=0.95)
    assert slope == res[0]
    assert intercept == res[1]

    # Test case 3: Test with alpha = 2 and identical observations
    x = np.array([10, 10, 10, 10, 10])
    y = np.array([2, 4, 6, 8, 10])
    alpha = 2
    slope, intercept = linear_fit(x, y, alpha)
    with pytest.raises(ValueError):
        res = linregress(x, y)
    assert slope == 0
    assert intercept == x[0]


def test_get_smooth_coeffs():
    # Test case 1: Both pct_hdd_k and pct_cdd_k are less than min_pct_k
    coeffs = get_smooth_coeffs(10, 0.005, 20, 0.005, min_pct_k=0.01)
    assert np.allclose(coeffs, np.array([10, 0, 20, 0]))

    # Test case 2: pct_hdd_k is greater than min_pct_k and pct_cdd_k is less than min_pct_k
    coeffs = get_smooth_coeffs(10, 0.1, 20, 0.005, min_pct_k=0.01)
    assert np.allclose(coeffs, np.array([11, 1, 19.95, 0.05]))

    # Test case 3: pct_cdd_k is greater than min_pct_k and pct_hdd_k is less than min_pct_k
    coeffs = get_smooth_coeffs(10, 0.005, 20, 0.1, min_pct_k=0.01)
    assert np.allclose(coeffs, np.array([10.05, 0.05, 19, 1]))

    # Test case 4: pct_hdd_k and pct_cdd_k are both greater than min_pct_k and sum to less than or equal to 1
    coeffs = get_smooth_coeffs(10, 0.1, 20, 0.2, min_pct_k=0.01)
    assert np.allclose(coeffs, np.array([11, 1, 18, 2]))

    # Test case 5: pct_hdd_k and pct_cdd_k are both greater than min_pct_k and sum to greater than 1
    coeffs = get_smooth_coeffs(10, 0.5, 20, 0.6, min_pct_k=0.01)
    assert np.allclose(
        coeffs, np.array([14.54545455, 4.54545455, 14.54545455, 5.45454545])
    )

    # Test case 6: pct_match is 1.0, so the smoothed curve should converge at - or + inf
    coeffs = get_smooth_coeffs(10, 0.1, 20, 0.2, min_pct_k=0.01)
    assert np.allclose(coeffs, np.array([11, 1, 18, 2]))


def test_fix_identical_bnds():
    # Test case 1: No bounds are identical
    bnds = np.array([[1, 2], [3, 4], [5, 6]])
    assert np.array_equal(fix_identical_bnds(bnds), bnds)

    # Test case 2: One bound is identical
    bnds = np.array([[1, 2], [3, 3], [5, 6]])
    expected_output = np.array([[1, 2], [2, 4], [5, 6]])
    assert np.array_equal(fix_identical_bnds(bnds), expected_output)

    # Test case 3: Multiple bounds are identical
    bnds = np.array([[1, 2], [3, 3], [5, 5]])
    expected_output = np.array([[1, 2], [2, 4], [4, 6]])
    assert np.array_equal(fix_identical_bnds(bnds), expected_output)

    # Test case 4: All bounds are identical
    bnds = np.array([[1, 1], [1, 1], [1, 1]])
    expected_output = np.array([[0, 2], [0, 2], [0, 2]])
    assert np.array_equal(fix_identical_bnds(bnds), expected_output)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
import logging
from typing import Union
import pydantic

__all__ = ("EEMeterWarning",)


class EEMeterWarning(pydantic.BaseModel):
    """An object representing a warning and data associated with it.

    Attributes
    ----------
    qualified_name : :any:`str`
        Qualified name, e.g., `'eemeter.method_abc.missing_data'`.
    description : :any:`str`
        Prose describing the nature of the warning.
    data : :any:`dict`
        Data that reproducibly shows why the warning was issued. Data should
        be JSON serializable.
    """

    qualified_name: str
    description: str
    data: Union[dict, list]

    def __repr__(self):
        return "EEMeterWarning(qualified_name={})".format(self.qualified_name)

    def __str__(self):
        return repr(self)

    def json(self) -> dict:
        """Return a JSON-serializable representation of this result.

        The output of this function can be converted to a serialized string
        with :any:`json.dumps`.
        """
        return {
            "qualified_name": self.qualified_name,
            "description": self.description,
            "data": self.data,
        }

    def warn(self):
        data = ""
        if self.data:
            data = f"\n{self.data}"
        logging.getLogger("eemeter").warning(f"{self.description}{data}")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
from typing import Optional, Union

import numpy as np
import pandas as pd

from opendsm.eemeter.common.data_processor_utilities import compute_minimum_granularity
from opendsm.eemeter.common.features import compute_temperature_features, merge_features
from opendsm.eemeter.models.hourly_caltrack.usage_per_day import (
    caltrack_sufficiency_criteria,
)


class HourlyReportingData:
    def __init__(self, df: pd.DataFrame, is_electricity_data: bool):
        if "observed" not in df.columns:
            df["observed"] = np.nan

        if is_electricity_data:
            df.loc[df["observed"] == 0, "observed"] = np.nan

        df = self._correct_frequency(df)

        self.df = df
        self.warnings = []
        self.disqualification = []

    def _correct_frequency(self, df: pd.DataFrame):
        meter = df["observed"]
        temp = df["temperature"]

        # unknown for weirdly large frequencies. Anything higher frequency than hourly frequency still comes up as hourly
        min_granularity = compute_minimum_granularity(meter.dropna().index, "unknown")

        if meter.index.inferred_freq is None and min_granularity != "hourly":
            raise ValueError(
                f"Meter Data must be atleast hourly, but is {min_granularity}."
            )
        else:
            # TODO : Add the high frequency check for meter data
            meter = meter.resample("h").sum(min_count=1)
            meter.index.freq = "h"

        # TODO : Add the high frequency check for temperature data and add NaNs
        temp = temp.resample("h").mean()
        temp.index.freq = "h"

        return merge_features([meter, temp], keep_partial_nan_rows=True)

    @classmethod
    def from_series(
        cls,
        meter_data: Optional[pd.Series],
        temperature_data: pd.Series,
        is_electricity_data: bool,
    ):
        # TODO verify
        if meter_data is None:
            meter_data = temperature_data.copy().rename("observed") * np.nan
        df = merge_features([meter_data, temperature_data], keep_partial_nan_rows=True)
        df = df.rename(
            {
                df.columns[0]: "observed",
                df.columns[1]: "temperature",
            },
            axis=1,
        )
        return cls(df, is_electricity_data)


class HourlyBaselineData(HourlyReportingData):
    def __init__(self, df: pd.DataFrame, is_electricity_data: bool):
        if is_electricity_data:
            df.loc[df["observed"] == 0, "observed"] = np.nan

        df = self._correct_frequency(df)

        self.df = df
        self.warnings = self._check_data_sufficiency()
        self.disqualification = []

    def _check_data_sufficiency(self):
        meter = self.df["observed"].rename("meter_value")
        temp = self.df["temperature"]

        temperature_features = compute_temperature_features(
            meter.index,
            temp,
            data_quality=True,
        )

        sufficiency_df = merge_features([meter, temperature_features])
        sufficiency = caltrack_sufficiency_criteria(
            sufficiency_df, requested_start=None, requested_end=None
        )
        return sufficiency.warnings

    @classmethod
    def from_series(
        cls,
        meter_data: Union[pd.Series, pd.DataFrame],
        temperature_data: Union[pd.Series, pd.DataFrame],
        is_electricity_data: bool,
    ):
        if isinstance(meter_data, pd.Series):
            meter_data = meter_data.to_frame()
        if isinstance(temperature_data, pd.Series):
            temperature_data = temperature_data.to_frame()
        meter_data = meter_data.rename(columns={meter_data.columns[0]: "observed"})
        temperature_data = temperature_data.rename(
            columns={temperature_data.columns[0]: "temperature"}
        )
        temperature_data.index = temperature_data.index.tz_convert(
            meter_data.index.tzinfo
        )
        df = pd.concat([meter_data, temperature_data], axis=1).dropna()
        return cls(df, is_electricity_data)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
import numpy as np
from scipy.linalg import eigh
from scipy.ndimage import median_filter
from scipy.optimize import minimize_scalar


def ellipsoid_intersection_test(mu_A, mu_B, cov_A, cov_B):
    """
    Tests whether two ellipsoids intersect or not. The ellipsoids are defined by their mean vectors and covariance matrices.
    The function uses the K-function to calculate the intersection of the ellipsoids. If the K-function is greater than or equal to 0,
    then the ellipsoids intersect, otherwise they do not.

    Parameters:
    mu_A (numpy.ndarray): Mean vector of the first ellipsoid.
    mu_B (numpy.ndarray): Mean vector of the second ellipsoid.
    cov_A (numpy.ndarray): Covariance matrix of the first ellipsoid.
    cov_B (numpy.ndarray): Covariance matrix of the second ellipsoid.

    Returns:
    bool: True if the ellipsoids intersect, False otherwise.
    """

    # Fix if all values are the same in 1 direction, "brent" doesn't work well with this
    if cov_A[1, 1] == 0:
        cov_A[1, 1] = 1e-14

    if cov_B[1, 1] == 0:
        cov_B[1, 1] = 1e-14

    lambdas, phi = eigh(cov_A, b=cov_B)
    v_squared = np.dot(phi.T, mu_A - mu_B) ** 2

    res = minimize_scalar(
        ellipsoid_K_function,
        #   bracket = [0.0, 0.5, 1.0],
        bounds=[0.0, 1.0],
        args=(lambdas, v_squared),
        method="bounded",
    )

    if res.fun[0] >= 0:
        return True
    return False


def ellipsoid_K_function(ss, lambdas, v_squared):
    """
    The K-function is a measure of spatial point pattern, often used in spatial statistics
    to analyze the clustering or dispersion of points in a dataset. The formula used in this
    code is a specific calculation for an ellipsoid.

    Parameters:
    ss (float): A scalar value between 0 and 1.
    lambdas (numpy.ndarray): A 1D numpy array of eigenvalues of the covariance matrix.
    v_squared (numpy.ndarray): A 1D numpy array of squared differences between the means of two ellipsoids.

    Returns:
    float: The value of the K-function for the given input values.
    """
    ss = np.array(ss).reshape((-1, 1))
    lambdas = np.array(lambdas).reshape((1, -1))
    v_squared = np.array(v_squared).reshape((1, -1))

    return 1 - np.sum(v_squared * ((ss * (1 - ss)) / (1 + ss * (lambdas - 1))), axis=1)


def confidence_ellipse(x, y, var=np.ones([2, 2]) * 1.96):
    """
    Compute the confidence ellipse for a 2D dataset.

    Parameters:
    x (numpy.ndarray): The x-coordinates of the data points.
    y (numpy.ndarray): The y-coordinates of the data points.
    var (numpy.ndarray): The variance of the data points. Default is 1.96.

    Returns:
    list: A list containing the mean, covariance, major and minor axis lengths, and rotation angle of the ellipse.

    """

    # Applying a median filter to help with outliers
    idx_sorted = np.argsort(x).flatten()
    idx_original = np.argsort(idx_sorted).flatten()

    # size could be changed with justification
    y = median_filter(y[idx_sorted], size=5)[idx_original]

    # Computing the covariance and ellipse parameter values
    cov = np.cov(x, y) * var  # scale covariances by std choice
    ab_sqr, v = np.linalg.eig(cov)
    [a, b] = np.sqrt(ab_sqr)
    phi = np.arctan2(*v[:, 0][::-1])

    mu = np.array([np.mean(x), np.mean(y)])

    return mu, cov, a, b, phi


def robust_confidence_ellipse(x, y, var=np.ones([2, 2]) * 1.96, outlier_std=3, N=3):
    """
    Computes a robust confidence ellipse for a set of points.

    Parameters:
    x (numpy.ndarray): Array of x-coordinates of the points.
    y (numpy.ndarray): Array of y-coordinates of the points.
    var (numpy.ndarray): Variance-covariance matrix. Default is a 2x2 matrix with 1.96 in the diagonal.
    outlier_std (float): Standard deviation for outlier detection. Default is 3.
    N (int): Number of iterations for outlier removal. Default is 3.

    Returns:
    list: A list containing the mean, covariance matrix, major and minor axis lengths, and rotation angle of the ellipse.
    """

    var_outlier = np.ones([2, 2]) * outlier_std**2

    # remove outliers in N iterations
    for n in range(N):
        if len(x) <= 1 or np.all(x == x[0]) or np.all(y == y[0]):
            break

        mu, cov, a, b, phi = confidence_ellipse(x, y, var_outlier)

        if a == 0 or b == 0:
            break

        # Center points
        xc = x - mu[0]
        yc = y - mu[1]

        # Rotate points so ellipse is aligned with axes
        phi *= -1
        xct = xc * np.cos(phi) - yc * np.sin(phi)
        yct = xc * np.sin(phi) + yc * np.cos(phi)

        # normalize to a circle of radius 1
        r = (xct / a) ** 2 + (yct / b) ** 2

        idx = np.argwhere(r <= 1).flatten()  # non-outlier points

        # if all outliers, break
        if len(idx) < 3:
            break

        # if no outliers, break
        if len(idx) == len(x):
            break

        x = x[idx]
        y = y[idx]

    if (len(x) < 3) or np.all(x == x[0]) or np.all(y == y[0]):
        mu = cov = a = b = phi = None
        return [mu, cov, a, b, phi]

    return confidence_ellipse(x, y, var)


def ellipsoid_split_filter(meter, n_std=[1.4, 1.4]):
    """
    Filters a set of points based on a robust confidence ellipse. The points are split into groups using robust ellipses computed
    and then tested for intersection. This determines whether separate keys are needed for different seasons and day types.

    Parameters:
    meter (pandas.DataFrame): Dataframe containing the points to be filtered.
    n_std (float or list): Standard deviation for outlier detection. Default is [1.4, 1.4].

    Returns:
    dict: A dictionary containing the filtered points for each season and day type.
    """

    if isinstance(n_std, float):
        var = np.ones([2, 2]) * n_std**2
    else:
        std = np.array(n_std)[:, None]
        var = std.T * std

    cluster_ellipse = {}
    for season in ["summer", "shoulder", "winter"]:
        for day_type, day_num in enumerate([[1, 2, 3, 4, 5], [6, 7]]):
            if day_type == 0:
                key = f"wd-{season[:2]}"
            else:
                key = f"we-{season[:2]}"

            meter_season = meter[
                (meter["season"] == season) & (meter["observed"].notna())
            ]
            meter_season = meter_season[meter_season["day_of_week"].isin(day_num)]
            meter_season = meter_season.sort_values(by=["temperature"])

            T = meter_season["temperature"].values
            obs = meter_season["observed"].values

            if (len(T) < 3) or (len(obs) < 3):
                mu = cov = a = b = phi = None
            else:
                mu, cov, a, b, phi = robust_confidence_ellipse(
                    T, obs, var, outlier_std=3.6, N=3
                )
                # mu, cov, a, b, phi = confidence_ellipse(T, obs, std_sqr)

            cluster_ellipse[key] = {"mu": mu, "cov": cov, "a": a, "b": b, "phi": phi}

    combos = {
        "summer": [
            [["wd-su", "wd-sh"], ["we-su", "we-sh"]],
            [["wd-su", "wd-wi"], ["we-su", "we-wi"]],
        ],
        "shoulder": [
            [["wd-su", "wd-sh"], ["we-su", "we-sh"]],
            [["wd-sh", "wd-wi"], ["we-sh", "we-wi"]],
        ],
        "winter": [
            [["wd-sh", "wd-wi"], ["we-sh", "we-wi"]],
            [["wd-su", "wd-wi"], ["we-su", "we-wi"]],
        ],
        "weekday_weekend": [
            [["wd-su", "we-su"], ["wd-sh", "we-sh"], ["wd-wi", "we-wi"]]
        ],
    }

    ellipse_overlap = {}
    allow_separate = {
        "summer": [False, False],
        "shoulder": [False, False],
        "winter": [False, False],
        "weekday_weekend": [False],
    }
    for key in allow_separate.keys():
        for i, season_wd_we in enumerate(combos[key]):
            for combo in season_wd_we:
                combo_str = "__".join(combo)

                if combo_str not in ellipse_overlap:
                    mu_A = cluster_ellipse[combo[0]]["mu"]
                    cov_A = cluster_ellipse[combo[0]]["cov"]
                    mu_B = cluster_ellipse[combo[1]]["mu"]
                    cov_B = cluster_ellipse[combo[1]]["cov"]

                    if all([coef is not None for coef in [mu_A, mu_B, cov_A, cov_B]]):
                        ellipse_overlap[combo_str] = ellipsoid_intersection_test(
                            mu_A, mu_B, cov_A, cov_B
                        )
                    else:
                        ellipse_overlap[combo_str] = False

                if not ellipse_overlap[combo_str]:
                    allow_separate[key][i] = True
                    break

        allow_separate[key] = all(allow_separate[key])

    return allow_separate



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
import numpy as np
from opendsm.eemeter.models.daily.parameters import ModelCoefficients
from opendsm.eemeter.models.daily.parameters import ModelType
from opendsm.eemeter.models.daily.utilities.settings import DailySettings as Settings
from opendsm.eemeter.models.daily.base_models.c_hdd_tidd import fit_c_hdd_tidd
from opendsm.eemeter.models.daily.fit_base_models import _get_opt_settings


def test_fit_c_hdd_tidd_smooth():
    # Test case 1: Test with initial_fit=True
    T = np.array([10, 20, 30, 40, 50, 60, 70, 80, 90, 100]).astype(float)
    obs = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10]).astype(float)
    settings = Settings(
        developer_mode=True,
        alpha_selection=0.1,
        alpha_final=0.2,
        segment_minimum_count=5,
        maximum_slope_OoM_scaler=1,
    )
    opt_options = _get_opt_settings(settings)
    x0 = ModelCoefficients(
        model_type=ModelType.HDD_TIDD_SMOOTH,
        intercept=0.0,
        hdd_bp=0.0,
        hdd_beta=0.0,
        hdd_k=0.0,
        cdd_bp=None,
        cdd_beta=None,
        cdd_k=None,
    )
    bnds = None
    weights = None
    initial_fit = True
    smooth = True
    res = fit_c_hdd_tidd(T, obs, weights, settings, opt_options, smooth, x0, bnds, initial_fit)
    assert res.success == True

    # Test case 2: Test with initial_fit=False
    T = np.array([10, 20, 30, 40, 50, 60, 70, 80, 90, 100]).astype(float)
    obs = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10]).astype(float)
    settings = Settings(
        developer_mode=True,
        alpha_selection=0.1,
        alpha_final=0.2,
        segment_minimum_count=5,
        maximum_slope_OoM_scaler=1,
    )
    opt_options = _get_opt_settings(settings)
    x0 = ModelCoefficients(
        model_type=ModelType.HDD_TIDD_SMOOTH,
        intercept=0.0,
        hdd_bp=0.0,
        hdd_beta=0.0,
        hdd_k=0.0,
        cdd_bp=None,
        cdd_beta=None,
        cdd_k=None,
    )
    bnds = None
    initial_fit = False
    smooth = True
    res = fit_c_hdd_tidd(T, obs, weights, settings, opt_options, smooth, x0, bnds, initial_fit)
    assert res.success == True

    # Test case 3: Test with x0=None
    T = np.array([10, 20, 30, 40, 50, 60, 70, 80, 90, 100]).astype(float)
    obs = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10]).astype(float)
    settings = Settings(
        developer_mode=True,
        alpha_selection=0.1,
        alpha_final=0.2,
        segment_minimum_count=5,
        maximum_slope_OoM_scaler=1,
    )
    opt_options = _get_opt_settings(settings)
    x0 = None
    bnds = None
    initial_fit = True
    smooth = True
    res = fit_c_hdd_tidd(T, obs, weights, settings, opt_options, smooth, x0, bnds, initial_fit)
    assert res.success == True